from mathutils import Euler
from mathutils import Matrix
from mathutils import Vector
from uplogic.nodes import ULParameterNode


class ULEulerToMatrix(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.input_e = None
        self.matrix = Matrix()
        self.OUT = self.add_output(self.get_matrix)

    def get_matrix(self):
        vec = self.get_input(self.input_e)
        if isinstance(vec, Vector):
            vec = Euler((vec.x, vec.y, vec.z), 'XYZ')
        return vec.to_matrix()



from uplogic.nodes import ULParameterNode


class ULGetParent(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.OUT = self.add_output(self.get_parent)

    def get_parent(self):
        return self.get_input(self.game_object).parent



from uplogic.nodes import ULParameterNode
from uplogic.input import VRHeadset


class ULGetVRHeadsetValues(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.controller = VRHeadset()
        self.POS = self.add_output(self.get_pos)
        self.ORI = self.add_output(self.get_ori)

    def get_pos(self):
        return self.controller.position
    
    def get_ori(self):
        return self.controller.orientation.to_euler()



from .fmodloadbank import FModLoadBankNode
from .fmodstartevent import FModStartEventNode
from .fmodgeteventattribute import FModGetEventAttributeNode
from .fmodseteventattribute import FModSetEventAttributeNode
from .fmodgeteventparameter import FModGetEventParameterNode
from .fmodseteventparameter import FModSetEventParameterNode
from .fmodmodifyeventparameter import FModModifyEventParameterNode



from uplogic.nodes import ULParameterNode
import math


class ULAbsoluteValue(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.out_value = False
        self.OUT = self.add_output(self._get_out_value)

    def _get_out_value(self):
        value = self.get_input(self.value)
        return math.fabs(value)



from uplogic.nodes import ULConditionNode


class ULSensorPositive(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.obj_name = None
        self.sens_name = None
        self.OUT = self.add_output(self.get_sensor)

    def get_sensor(self):
        game_obj = self.get_input(self.obj_name)
        sens_name = self.get_input(self.sens_name)
        return game_obj.sensors[sens_name].positive



from mathutils import Vector
from uplogic.nodes import ULParameterNode


class ULVectorAbsolute(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.input_v = None
        self.output_v = Vector()
        self.OUTV = self.add_output(self.get_out_v)

    def get_out_v(self):
        vec = self.get_input(self.input_v)
        vec = vec.copy()
        vec.x = abs(vec.x)
        vec.y = abs(vec.y)
        vec.z = abs(vec.z)
        return vec



from uplogic.nodes import ULParameterNode


class ULWorldPosition(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.camera = None
        self.screen_x = None
        self.screen_y = None
        self.world_z = None
        self.OUT = self.add_output(self.get_pos)

    def get_pos(self):
        camera = self.get_input(self.camera)
        screen_x = self.get_input(self.screen_x)
        screen_y = self.get_input(self.screen_y)
        world_z = self.get_input(self.world_z)
        direction = camera.getScreenVect(screen_x, screen_y)
        origin = camera.worldPosition
        aim = direction * -world_z
        point = origin + (aim)
        return point



from uplogic.nodes import ULConditionNode
from uplogic.utils import is_invalid


class ULValueValid(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.checked_value = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        value = self.get_input(self.checked_value)
        return not is_invalid(value)



from bge.render import drawLine
from bge import logic
from mathutils import Vector
from bge.types import KX_GameObject
from bpy.types import Mesh


def draw_line(origin: Vector, target: Vector, color: list = [1, 1, 1, 1]):
    drawLine(
        origin,
        target,
        color
    )


def draw_arrow(origin: Vector, target: Vector, color: list = [1, 1, 1, 1]):
    cam = logic.getCurrentScene().active_camera
    target = Vector(target)
    origin = Vector(origin)

    cam_dir = target - cam.worldPosition

    direction = target - origin
    normal = direction.cross(cam_dir).normalized() * direction.length

    drawLine(
        target,
        target - direction * .2 + normal * .1,
        color
    )

    drawLine(
        target,
        target - direction * .2 - normal * .1,
        color
    )

    drawLine(
        origin,
        target,
        color
    )


def draw_path(points: list, color: list = [1, 1, 1, 1]):
    for i, p in enumerate(points):
        if i < len(points) - 1:
            drawLine(p, points[i+1], color)


def draw_arrow_path(points: list, color: list = [1, 1, 1, 1]):
    for i, p in enumerate(points):
        if i < len(points) - 1:
            draw_arrow(p, points[i+1], color)



def draw_cube(origin: Vector, width: float = 1, color: list = [1, 1, 1, 1], centered: bool = False):
    draw_box(origin, width, width, width, color, centered)


def draw_box(origin: Vector, width: float, length: float, height: float, color: list = [1, 1, 1, 1], centered: bool = False):
    is_obj = isinstance(origin, KX_GameObject)

    if is_obj:
        obj = origin
        origin = origin.worldPosition.copy()
        origin -= obj.worldPosition
        centered = True
    if centered:
        origin = origin.copy() - Vector((width * .5, length * .5, height * .5))

    c1: Vector = origin.copy()
    c2: Vector = origin.copy()
    c3: Vector = origin.copy()
    c4: Vector = origin.copy()
    c5: Vector = origin.copy()
    c6: Vector = origin.copy()
    c7: Vector = origin.copy()

    c1[0] += length

    c2[0] += length
    c2[1] += width

    c3[1] += width

    c4[2] += height

    c5[0] += length
    c5[2] += height

    c6[0] += length
    c6[1] += width
    c6[2] += height

    c7[1] += width
    c7[2] += height

    if is_obj:
        ori = obj.worldOrientation
        offset = obj.worldPosition
        origin = ori @ origin + offset
        c1 = ori @ c1 + offset
        c2 = ori @ c2 + offset
        c3 = ori @ c3 + offset
        c4 = ori @ c4 + offset
        c5 = ori @ c5 + offset
        c6 = ori @ c6 + offset
        c7 = ori @ c7 + offset

    drawLine(origin, c1, color)
    drawLine(c1, c2, color)
    drawLine(c2, c3, color)
    drawLine(c3, origin, color)

    drawLine(origin, c4, color)
    drawLine(c1, c5, color)
    drawLine(c2, c6, color)
    drawLine(c3, c7, color)

    drawLine(c4, c5, color)
    drawLine(c5, c6, color)
    drawLine(c6, c7, color)
    drawLine(c7, c4, color)


def draw_mesh(game_object: KX_GameObject, color: tuple = (1, 1, 1, 1)):
    mesh: Mesh = game_object.blenderObject.data
    for edge in mesh.edges:
        v1 = mesh.vertices[edge.vertices[0]]
        v2 = mesh.vertices[edge.vertices[1]]
        drawLine(
            game_object.worldTransform @ Vector(v1.co),
            game_object.worldTransform @ Vector(v2.co),
            color
        )


def draw_axis(game_object: KX_GameObject, length=1.0):
    xaxis = game_object.getAxisVect((1, 0, 0)) * length
    yaxis = game_object.getAxisVect((0, 1, 0)) * length
    zaxis = game_object.getAxisVect((0, 0, 1)) * length
    wpos = game_object.worldPosition.copy()
    drawLine(wpos, wpos + xaxis, (1, 0, 0))  # +X
    drawLine(wpos, wpos + yaxis, (0, 1, 0))  # +Y
    drawLine(wpos, wpos + zaxis, (0, 0, 1))  # +Z
    drawLine(wpos, wpos - xaxis, (.5, 0, 0))  # -X
    drawLine(wpos, wpos - yaxis, (0, .5, 0))  # -Y
    drawLine(wpos, wpos - zaxis, (0, 0, .5))  # -Z



from uplogic.nodes import ULActionNode
from uplogic.utils.objects import move_to


class ULMoveTo(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.moving_object = None
        self.destination_point = None
        self.speed = None
        self.dynamic = None
        self.distance = None
        self._reached = False
        self.OUT = self.add_output(self.get_done)
        self.REACHED = self.add_output(self.get_reached)

    def get_done(self):
        return self._done

    def get_reached(self):
        return self._reached

    def evaluate(self):
        if not self.get_condition():
            return
        moving_object = self.get_input(self.moving_object)
        destination_point = self.get_input(self.destination_point)
        speed = self.get_input(self.speed)
        distance = self.get_input(self.distance)
        dynamic = self.get_input(self.dynamic)
        self._reached = move_to(
            moving_object,
            destination_point,
            speed,
            distance
        )
        self._done = True



from bge import logic
from uplogic.nodes import ULConditionNode


class ULGamepadButtonUp(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.pulse = False
        self.button = 0
        self.index = None
        self._last_value = False
        self._up_value = None
        self.BUTTON = self.add_output(self.get_button)
        self.initialized = False

    def get_button(self):
        pressed = False
        index = self.get_input(self.index)
        if logic.joysticks[index]:
            joystick = logic.joysticks[index]
        else:
            return False

        button_down = (
            True if
            self.button in joystick.activeButtons else
            False
        )

        if button_down != self._last_value and not button_down:
            if not self.pulse and not self.initialized:
                self.initialized = True
            pressed = True
        elif (pressed and self.initialized) or button_down:

            pressed = False
            self.initialized = False
        elif not (self.initialized and button_down) and pressed:
            pressed = True

        self._last_value = button_down
        return pressed



import bge, bpy


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


# class Engine(object):

#     _pre_draw = None
#     _post_draw = None
#     _pre_draw_setup = None

#     @classproperty
#     def bge_scene(cls) -> bge.types.KX_Scene:
#         bge.logic.getCurrentScene()
    
#     @classproperty
#     def bpy_scene(cls):
#         return bpy.data.scenes[cls.bge_scene.name]

#     @classproperty
#     def pre_draw(cls):
#         scene = cls.bpy_scene
#         return (
#             cls.bpy_scene.pre_draw
#             if scene.game_settings.use_viewport_renderer else
#             cls.bge_scene.pre_draw
#         )

#     @pre_draw.setter
#     def pre_draw(cls, value):
#         return

#     @classproperty
#     def post_draw(cls):
#         return cls._bar

#     @post_draw.setter
#     def post_draw(cls, value):
#         return

#     @classproperty
#     def pre_draw_setup(cls):
#         return cls._bar

#     @pre_draw_setup.setter
#     def pre_draw_setup(cls, value):
#         return


# def pre_draw(cb):
#     print(cb.__self__)
#     scene = bge.logic.getCurrentScene()
#     scene.pre_draw.append(cb)
    # scene = Engine.bpy_scene
    # return (
    #     Engine.bpy_scene.pre_draw
    #     if scene.game_settings.use_viewport_renderer else
    #     Engine.bge_scene.pre_draw



from .widget import Widget
from .path import Path
from uplogic.utils.scene import world_to_screen
from bge.render import getWindowWidth, getWindowHeight
from bge.types import KX_GameObject
from mathutils import Vector
from gpu_extras.batch import batch_for_shader
import gpu


class WorldPath(Path):

    def __init__(
        self,
        points=[],
        line_width=1,
        line_color=(1.0, 1.0, 1.0, 1.0)
    ):
        Widget.__init__(self, (0, 0), (0, 0), (0, 0, 0, 0))
        self.points = points
        self.line_color = line_color
        self.line_width = line_width
        self.start()
    
    def _build_shader(self):
        if self.parent is None:
            return
        points = []
        for point in self.points:
            point = world_to_screen(point)
            point *= Vector((
                getWindowWidth(),
                getWindowHeight()
            ))
            points.append(point)
        vertices = self._vertices = points
        self._shader = gpu.shader.from_builtin('UNIFORM_COLOR')
        self._batch_line = batch_for_shader(self._shader, 'LINE_STRIP', {"pos": vertices})


class ObjectPath(Path):
    def __init__(
        self,
        object=None,
        line_width=1,
        line_color=(1.0, 1.0, 1.0, 1.0)
    ):
        Widget.__init__(self, (0, 0), (0, 0), (0, 0, 0, 0))
        self.object: KX_GameObject = object
        self.line_color = line_color
        self.line_width = line_width
        self.start()

    def _build_shader(self):
        if self.parent is None:
            return
        points = []
        mesh = self.object.blenderObject.data
        for edge in mesh.edges:
            v1 = mesh.vertices[edge.vertices[0]]
            v2 = mesh.vertices[edge.vertices[1]]
            v1 = world_to_screen(self.object.worldTransform @ Vector(v1.co))
            v2 = world_to_screen(self.object.worldTransform @ Vector(v2.co))
            # # point = world_to_screen(point)
            v1 *= Vector((
                getWindowWidth(),
                getWindowHeight()
            ))
            v2 *= Vector((
                getWindowWidth(),
                getWindowHeight()
            ))
            points.append(v1)
            points.append(v2)
        vertices = self._vertices = points
        self._shader = gpu.shader.from_builtin('UNIFORM_COLOR')
        self._batch_line = batch_for_shader(self._shader, 'LINES', {"pos": vertices})



from uplogic.nodes import ULActionNode
from uplogic.network import Server
from uplogic.events import send, receive


class ULLocalServer(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.start_cond = None
        self.ip_address = None
        self.port = None
        self.stop_cond = None

        self.on_init = False

        self._started = False
        self._stopped = False
        self._was_running = False
        self._received = False
        self._message = None

        self.STARTED = self.add_output(self.get_started)
        self.RUNNING = self.add_output(self.get_running)
        self.STOPPED = self.add_output(self.get_stopped)
        self.SERVER = self.add_output(self.get_server)
        self.RECEIVED = self.add_output(self.get_received)
        self.MSG = self.add_output(self.get_message)

    def get_started(self):
        return self._started

    def get_running(self):
        return self._server.running

    def get_stopped(self):
        return self._stopped

    def get_server(self):
        return self._server

    def get_received(self):
        return self._received

    def get_message(self):
        return self._received.content if self._received else None

    def on_receive(self, msg):
        send(self._server, msg)

    def setup(self, tree):
        self._server = Server(
            self.get_input(self.ip_address),
            self.get_input(self.port),
            self.on_init
        )
        self._server.on_receive = self.on_receive

    def evaluate(self):
        self._started = False
        self._stopped = False
        self._received = receive(self._server)
        if self._was_running != self._server.running:
            if self._server.running:
                self._started = True
            else:
                self._stopped = True
            self._was_running = self._server.running
        if self.get_input(self.start_cond):
            self._server.start()
        if self.get_input(self.stop_cond):
            self._server.shutdown()
        if not self._server.running:
            self._server.ip = self.get_input(self.ip_address)
            self._server.port = self.get_input(self.port)



from uplogic.nodes import ULActionNode
from uplogic.utils.math import clamp
from uplogic.utils.constants import PROP_TYPE_INVALID


class ULModifyProperty(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.property_name = None
        self.property_value = None
        self.operator = None
        self.mode = 0
        self.clamp = False
        self.min_value = 0.0
        self.max_value = 1.0
        self.OUT = self.add_output(self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object = self.get_input(self.game_object)
        property_name = self.get_input(self.property_name)
        property_value = self.get_input(self.property_value)
        obj = game_object.blenderObject if self.mode else game_object
        value = obj.get(property_name, PROP_TYPE_INVALID)
        if value is PROP_TYPE_INVALID:
            return
        result = self.operator(value, property_value)
        if self.clamp:
            result = clamp(
                result,
                self.get_input(self.min_value),
                self.get_input(self.max_value)
            )
        obj[property_name] = result
        if self.mode:
            obj.color = obj.color
        self._done = True



from uplogic.nodes import ULActionNode
from bge.types import KX_GameObject


class ULSetLightShadow(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.lamp = None
        self.use_shadow = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        light: KX_GameObject = self.get_input(self.lamp)
        use_shadow = self.get_input(self.use_shadow)
        light = light.blenderObject.data
        light.use_shadow = use_shadow
        self._done = True



from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.utils.raycasting import raycast_camera, RayCastCameraData
from uplogic.utils import get_bitmask


class ULCameraRayCast(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.camera = None
        self.aim = None
        self.property_name = None
        self.xray = None
        self.mask = get_bitmask(all=True)
        self.distance = None
        self._data = RayCastCameraData((None, None, None))
        self.RESULT = self.add_output(self.get_result)
        self.PICKED_OBJECT = self.add_output(self.get_picked_object)
        self.PICKED_POINT = self.add_output(self.get_picked_point)
        self.PICKED_NORMAL = self.add_output(self.get_picked_normal)

    def get_result(self):
        return self._data.obj

    def get_picked_object(self):
        return self._data.obj

    def get_picked_point(self):
        return self._data.point

    def get_picked_normal(self):
        return self._data.normal

    def evaluate(self):
        if not self.get_condition():
            self._data = RayCastCameraData((None, None, None))
            return
        self._data = raycast_camera(
            distance=self.get_input(self.distance),
            prop=self.get_input(self.property_name),
            xray=self.get_input(self.xray),
            aim=self.get_input(self.aim),
            mask=self.get_input(self.mask)
        )



from uplogic.nodes import ULActionNode
from uplogic.shaders import FXAA
from uplogic.shaders import HBAO
from uplogic.shaders import SSAO
from uplogic.shaders import Vignette
from uplogic.shaders import Brightness
from uplogic.shaders import ChromaticAberration
from uplogic.shaders import Grayscale
from uplogic.shaders import Levels
from uplogic.shaders import Mist
from uplogic.shaders import Blur


class ULAddFilter(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.pass_idx = None
        self.brightness = None
        self.power = None
        self.density = None
        self.color = None
        self.start = None
        self.end = None
        self.filter = None
        self.filter_type = 'FXAA'
        self.OUT = self.add_output(self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        condition = self.get_condition()
        ftype = self.filter_type
        if not condition or self.filter:
            if self.filter:
                if ftype in [3, 'VIGNETTE']:
                    self.filter.uniforms['power'] = self.get_input(self.power)
                    self.filter.uniforms['color'] = self.get_input(self.color)
                elif ftype in [4, 'BRIGHTNESS']:
                    self.filter.uniforms['brightness'] = self.get_input(self.brightness)
                elif ftype in [5, 'CHROMAB']:
                    self.filter.uniforms['power'] = self.get_input(self.power)
                elif ftype in [6, 'GRAYSCALE']:
                    self.filter.uniforms['power'] = self.get_input(self.power)
                elif ftype in [7, 'LEVELS']:
                    self.filter.uniforms['color'] = self.get_input(self.color)
                elif ftype in [8, 'MIST']:
                    self.filter.uniforms['power'] = self.get_input(self.power)
                    self.filter.uniforms['color'] = self.get_input(self.color)
                    self.filter.uniforms['start'] = self.get_input(self.start)
                    self.filter.uniforms['density'] = self.get_input(self.density)
                elif ftype in [1, 2, 'HBAO', 'SSAO']:
                    self.filter.uniforms['power'] = self.get_input(self.power)
                elif ftype in [9, 'BLUR']:
                    self.filter.uniforms['power'] = self.get_input(self.power)
            return

        if ftype in [0, 'FXAA']:
            self.filter = FXAA(self.get_input(self.pass_idx))
        elif ftype in [1, 'HBAO']:
            self.filter = HBAO(self.get_input(self.power), self.get_input(self.pass_idx))
        elif ftype in [2, 'SSAO']:
            self.filter = SSAO(self.get_input(self.power), self.get_input(self.pass_idx))
        elif ftype in [3, 'VIGNETTE']:
            self.filter = Vignette(
                self.get_input(self.power),
                self.get_input(self.color),
                self.get_input(self.pass_idx)
            )
        elif ftype in [4, 'BRIGHTNESS']:
            self.filter = Brightness(
                self.get_input(self.brightness),
                self.get_input(self.pass_idx)
            )
        elif ftype in [5, 'CHROMAB']:
            self.filter = ChromaticAberration(
                self.get_input(self.power),
                self.get_input(self.pass_idx)
            )
        elif ftype in [6, 'GRAYSCALE']:
            self.filter = Grayscale(
                self.get_input(self.power),
                self.get_input(self.pass_idx)
            )
        elif ftype in [7, 'LEVELS']:
            self.filter = Levels(
                self.get_input(self.color),
                self.get_input(self.pass_idx)
            )
        elif ftype in [8, 'MIST']:
            self.filter = Mist(
                self.get_input(self.start),
                self.get_input(self.density),
                self.get_input(self.color),
                self.get_input(self.power),
                self.get_input(self.pass_idx)
            )
        elif ftype == 9:
            self.filter = Blur(
                16,
                self.get_input(self.power),
                self.get_input(self.pass_idx)
            )
        self._done = True



from mathutils import Vector
from uplogic.nodes import ULParameterNode


class ULGetCurvePoints(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.curve = None
        self.OUT = self.add_output(self.get_points)

    def get_points(self):
        obj = self.get_input(self.curve)
        offset = obj.worldPosition
        o = obj.blenderObject.data.splines[0]
        if o.type == 'BEZIER':
            return [Vector(p.co) + offset for p in o.bezier_points]
        return [Vector(p.co[:-1]) + offset for p in o.points]



from uplogic.nodes import ULActionNode
from uplogic.physics import RWD, Vehicle
from uplogic.utils.constants import VEHICLE


class ULVehicleApplyBraking(ULActionNode):
    def __init__(self, value_type=RWD):
        ULActionNode.__init__(self)
        self.value_type = str(value_type)
        self.condition = None
        self.vehicle = None
        self.wheelcount = None
        self._reset = False
        self.power = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object = self.get_input(self.vehicle)
        vehicle: Vehicle = game_object.get(VEHICLE, None)
        if vehicle is None:
            return
        value_type = self.get_input(self.value_type)
        wheelcount = self.get_input(self.wheelcount)
        power = self.get_input(self.power)
        vehicle.brake(power, value_type, wheelcount)
        self._done = True



from bge import constraints
from uplogic.nodes import ULActionNode


class ULSetCharacterGravity(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.gravity = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object = self.get_input(self.game_object)
        gravity = self.get_input(self.gravity)
        physics = constraints.getCharacter(game_object)
        if physics:
            physics.gravity = gravity
        else:
            game_object.gravity = gravity
        self._done = True



from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.utils.constants import LOGIC_OPERATORS
import math


class ULVectorAngleCheck(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.op: str = None
        self.vector: Vector = None
        self.vector_2: Vector = None
        self.value = None
        self._angle = 0
        self.OUT = self.add_output(self.get_done)
        self.ANGLE = self.add_output(self.get_angle)

    def get_angle(self):
        return self._angle

    def get_done(self):
        op: str = self.get_input(self.op)
        value: float = self.get_input(self.value)
        return LOGIC_OPERATORS[int(op)](self._angle, value)

    def evaluate(self):
        vector: Vector = self.get_input(self.vector)
        vector_2: Vector = self.get_input(self.vector_2)
        rad: float = vector.angle(vector_2)
        deg: float = rad * 180/math.pi
        self._angle = deg



from uplogic.nodes import ULActionNode
from uplogic.ui import Widget


class MoveWidgetNode(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.widget: Widget = None
        self.mode = '0'
        self.OUT = self.add_output(self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        if self.get_condition():
            widget: Widget = self.get_input(self.widget)
            operations = [
                widget.move_to_top,
                widget.move_up,
                widget.move_down,
                widget.move_to_bottom
            ]
            operations[int(self.mode)]()
            self._done = True



from uplogic.data import GlobalDB
from uplogic.nodes import ULParameterNode


class ULGetGlobalValue(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.data_id = None
        self.key = None
        self.default = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        data_id = self.get_input(self.data_id)
        key = self.get_input(self.key)
        default = self.get_input(self.default)
        db = GlobalDB.retrieve(data_id)
        return db.get(key, default)



from .audiosystem import AudioSystem  # noqa
from .audiosystem import set_master_volume  # noqa
from .audiosystem import set_lowpass  # noqa
from .audiosystem import set_vr_audio  # noqa
from .audiosystem import get_audio_system  # noqa
from .audiosystem import stop_all_audio  # noqa
from .sound import ULSound2D, Sound2D  # noqa
from .sound import Sample2D  # noqa
from .sound import Sample3D  # noqa
from .sound import ULSound3D, Sound3D  # noqa
from .sound import ULSpeaker2D, Speaker2D  # noqa
from .sound import ULSpeaker3D, Speaker3D  # noqa
from .sound import play_sound_2d, play_sound_3d  # noqa
from .music import ULMusic, Music  # noqa



from uplogic.nodes import ULParameterNode
from mathutils import Vector, Matrix
import bpy
from bge import logic


class ResizeVectorNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.vec_in = Vector((0, 0, 0))
        self.OUT = self.add_output(self.get_2D_vec)
        self._to_size = 3
        self.funcs = [
            self.get_2D_vec,
            self.get_3D_vec,
            self.get_4D_vec
        ]

    @property
    def to_size(self):
        return self._to_size

    @to_size.setter
    def to_size(self, val):
        self.OUT._value_getter = self.funcs[val]
        self._to_size = val

    def get_2D_vec(self):
        return self.get_input(self.vec_in).to_2d()

    def get_3D_vec(self):
        return self.get_input(self.vec_in).to_3d()

    def get_4D_vec(self):
        return self.get_input(self.vec_in).to_4d()



from bge.types import BL_ArmatureChannel, BL_ArmatureObject
from bge.logic import ROT_MODE_XYZ
from uplogic import console
from uplogic.utils.objects import GameObject
from mathutils import Vector, Quaternion, Matrix, Euler
from uplogic.utils.visualize import draw_line


class RigBone():

    def __init__(self, bone: BL_ArmatureChannel, armature: BL_ArmatureObject) -> None:
        self.world_space = False
        self.bone = bone
        self.armature = armature
        self._armature_data = self.armature.blenderObject.data
        self._pose = self.armature.blenderObject.pose

    @property
    def attr_transform(self):
        return self.armature.worldTransform if self.world_space else Matrix()

    @property
    def armature(self):
        return self._armature

    @armature.setter
    def armature(self, val):
        self._armature_data = val.blenderObject.data
        self._armature = val

    @property
    def name(self):
        return self.bone.name

    @name.setter
    def name(self, val):
        self._armature_data.blenderObject.bones[self.bone.name].name = val

    @property
    def constraints(self):
        return self._pose.bones[self.bone.name].constraints

    @property
    def location(self):
        return self._pose.bones[self.bone.name].location

    @location.setter
    def location(self, val):
        self._pose.bones[self.bone.name].location = val

    @property
    def pose_rotation_euler(self):
        bone = self._pose.bones[self.bone.name]
        _mode = bone.rotation_mode
        # bone.rotation_mode = ROT_MODE_XYZ
        bone.rotation_mode = 'XYZ'
        res = bone.rotation_euler
        bone.rotation_mode = _mode
        return res

    @pose_rotation_euler.setter
    def pose_rotation_euler(self, euler: Euler):
        bone = self._pose.bones[self.bone.name]
        _mode = bone.rotation_mode
        # bone.rotation_mode = ROT_MODE_XYZ
        bone.rotation_mode = 'XYZ'
        bone.rotation_euler = Euler(euler)
        self.armature.blenderObject.update_tag()
        bone.rotation_mode = _mode

    @property
    def head(self) -> Vector:
        return self.attr_transform @ self._armature_data.bones[self.bone.name].head

    @head.setter
    def head(self, value):
        self._armature_data.bones[self.bone.name].head = value

    @property
    def head_local(self) -> Vector:
        return self._armature_data.bones[self.bone.name].head_local

    @head_local.setter
    def head_local(self, value):
        self._armature_data.bones[self.bone.name].head_local = value

    @property
    def head_pose(self) -> Vector:
        return self.attr_transform @ self._pose.bones[self.bone.name].head

    @head_pose.setter
    def head_pose(self, val):
        self._pose.bones[self.bone.name].head = val

    @property
    def tail(self) -> Vector:
        return self.attr_transform @ self._armature_data.bones[self.bone.name].tail

    @tail.setter
    def tail(self, value):
        self._armature_data.bones[self.bone.name].tail = value

    @property
    def tail_local(self) -> Vector:
        return self._armature_data.bones[self.bone.name].tail_local

    @tail_local.setter
    def tail_local(self, value):
        self._armature_data.bones[self.bone.name].tail_local = value

    @property
    def tail_pose(self) -> Vector:
        return self.attr_transform @ self._pose.bones[self.bone.name].tail

    @tail_pose.setter
    def tail_pose(self, val):
        self._pose.bones[self.bone.name].tail = val

    @property
    def center(self) -> Vector:
        return self.head.lerp(self.tail, .5)

    @center.setter
    def center(self, val):
        console.debug('RigBone.center is read-only!')

    @property
    def center_local(self) -> Vector:
        return self.head_local.lerp(self.tail_local, .5)

    @center_local.setter
    def center_local(self, val):
        console.debug('RigBone.center_local is read-only!')

    @property
    def center_pose(self) -> Vector:
        return self.head_pose.lerp(self.tail_pose, .5)

    @center_pose.setter
    def center_pose(self, val):
        console.debug('RigBone.center_pose is read-only!')

    @property
    def inherit_rotation(self):
        return self._armature_data.bones[self.bone.name].use_inherit_rotation

    @inherit_rotation.setter
    def inherit_rotation(self, value):
        self._armature_data.bones[self.bone.name].use_inherit_rotation = value

    @property
    def inherit_scale(self):
        return self._armature_data.bones[self.bone.name].inherit_scale

    @inherit_scale.setter
    def inherit_scale(self, value):
        self._armature_data.bones[self.bone.name].inherit_scale = value

    @property
    def connected(self):
        return self._armature_data.bones[self.bone.name].use_connect

    @connected.setter
    def connected(self, value):
        self._armature_data.bones[self.bone.name].use_connect = value

    @property
    def deform(self):
        return self._armature_data.bones[self.bone.name].use_deform

    @deform.setter
    def deform(self, value):
        self._armature_data.bones[self.bone.name].use_deform = value

    @property
    def use_local_location(self):
        return self._armature_data.bones[self.bone.name].use_local_location

    @use_local_location.setter
    def use_local_location(self, value):
        self._armature_data.bones[self.bone.name].use_local_location = value

    @property
    def use_relative_parent(self):
        return self._armature_data.bones[self.bone.name].use_relative_parent

    @use_relative_parent.setter
    def use_relative_parent(self, value):
        self._armature_data.bones[self.bone.name].use_relative_parent = value

    @property
    def use_scale_easing(self):
        return self._armature_data.bones[self.bone.name].use_scale_easing

    @use_scale_easing.setter
    def use_scale_easing(self, value):
        self._armature_data.bones[self.bone.name].use_scale_easing = value

    @property
    def pose_matrix_cleaned(self):
        offset_m4 = (Matrix.Translation(self.bone.location) @ Quaternion(self.bone.rotation_quaternion).to_matrix().to_4x4())
        return self.bone.pose_matrix @ offset_m4.inverted()

    @property
    def localPosition(self):
        return self.bone.location @ self.bone.channel_matrix.inverted()

    @localPosition.setter
    def localPosition(self, val):
        self.bone.location = self.pose_matrix_cleaned.inverted() @ self.armature.worldTransform.inverted() @ (Vector(val) + self.armature.worldPosition)
        self.armature.blenderObject.update_tag()

    @property
    def worldPosition(self):
        return self.bone.location @ self.bone.pose_matrix.inverted()

    @worldPosition.setter
    def worldPosition(self, val):
        self.bone.location = self.pose_matrix_cleaned.inverted() @ self.armature.worldTransform.inverted() @ Vector(val)
        self.armature.blenderObject.update_tag()

    @property
    def worldOrientation(self):
        _mode = self.bone.rotation_mode
        self.bone.rotation_mode = ROT_MODE_XYZ
        res = self.bone.rotation_euler
        self.bone.rotation_mode = _mode
        return res

    @worldOrientation.setter
    def worldOrientation(self, val: Matrix):
        _mode = self.bone.rotation_mode
        self.bone.rotation_mode = ROT_MODE_XYZ
        self.bone.rotation_euler = Vector(val.to_euler())
        self.armature.blenderObject.update_tag()
        self.bone.rotation_mode = _mode


class RigBones(dict):
    pass


class Rig(GameObject):

    def __init__(
        self,
        armature
    ) -> None:
        self.game_object: BL_ArmatureObject = armature
        self.bones: RigBones[str, RigBone] = RigBones()
        for bone in self.game_object.channels:
            self.bones[bone.name] = RigBone(bone, armature)

    def set_bone_pos(self, bone: str, pos: Vector):
        self.bones[bone].worldPosition = pos

    def set_bone_rot(self, bone: str, rot: Quaternion):
        bone = self.game_object.channels[bone]
        bone.rotation_quaternion = bone.pose_matrix.inverted() @ Quaternion(rot)

    def visualize(self):
        for b in self.bones.values():
            st = b.world_space
            b.world_space = True
            # draw_line(b.head_pose, b.tail_pose, (.4, .88, .88))
            b.world_space = st



from uplogic.nodes import ULConditionNode
from bge.logic import getRealTime


class ULTimer(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.condition = None
        self.delta_time = None
        self._trigger = -1
        self.network = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        return self.result

    def setup(self, network):
        self.network = network

    def evaluate(self):
        self.result = False
        condition = self.get_input(self.condition)
        delta_time = self.get_input(self.delta_time)
        now = getRealTime()

        if condition:
            self._trigger = now + delta_time
        if self._trigger == -1 or now < self._trigger:
            self.result = False
        else:
            self.result = True
            self._trigger = -1



from uplogic.nodes import ULParameterNode
from uplogic.input import record_keyboard


class ULKeyLogger(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.only_characters = None
        self.PRESSED = self.add_output(self.get_pressed)
        self.CHARACTER = self.add_output(self.get_character)
        self.KEYCODE = self.add_output(self.get_keycode)

    def get_pressed(self):
        return len(self.scan()) > 0

    def get_keycode(self):
        return self.scan()[0].keycode

    def get_character(self):
        return self.scan()[0].character

    def scan(self):
        data = record_keyboard(False, not self.get_input(self.only_characters))
        return data

    def reset(self):
        super().reset()
        self._key_logged = False
        self._key_code = None
        self._character = None



from uplogic.data import GlobalDB
from uplogic.nodes import ULActionNode
from uplogic import console


class ULListGlobalValues(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.data_id = None
        self.print_d = None
        self.gv_dict = None
        self.OUT = self.add_output(self.get_done)
        self.VALUE = self.add_output(self.get_dict)

    def get_done(self):
        return self._done

    def get_dict(self):
        return self.gv_dict

    def evaluate(self):
        if not self.get_condition():
            return
        data_id = self.get_input(self.data_id)
        print_d = self.get_input(self.print_d)
        db = GlobalDB.retrieve(data_id)
        if print_d:
            console.log(f'[Logic Nodes] Global category "{data_id}":')
            for e in db.data:
                console.log('{}\t->\t{}'.format(e, db.data[e]))
            console.log('END ------------------------------------')
        self._done = True
        self.gv_dict = db.data



from uplogic.nodes import ULParameterNode


class ULKeyCode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.key_code = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        return self.get_input(self.key_code)



from uplogic.nodes import ULActionNode


class ULSetCurvePoints(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.curve_object = None
        self.points: list = None
        self._done: bool = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        curve_object = self.get_input(self.curve_object)
        points = self.get_input(self.points)
        if not points:
            return
        curve = curve_object.blenderObject.data
        for spline in curve.splines:
            curve.splines.remove(spline)
        spline = curve.splines.new('NURBS')
        pos = curve_object.worldPosition
        spline.points.add(len(points))
        for p, new_co in zip(spline.points, points):
            p.co = ([
                new_co.x - pos.x,
                new_co.y - pos.y,
                new_co.z - pos.z
            ] + [1.0])
        self._done = True



from uplogic.nodes import ULActionNode
from uplogic.audio.fmod import Event


class FModModifyEventParameterNode(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = False
        self.event = None
        self.clamp = False
        self.operation = None
        self.parameter = ''
        self.value = 0.0
        self.DONE = self.add_output(self.get_done)

    def evaluate(self):
        if not self.get_condition():
            return
        evt: Event = self.get_input(self.event)
        if evt is None:
            return
        param = evt.get_parameter(self.get_input(self.parameter))
        result = self.operation(param, self.get_input(self.value))
        evt.set_parameter(
            self.get_input(self.parameter),
            float(result)
        )



'''TODO: Documentation
'''

from os.path import isfile
from bge import logic
from bge.types import KX_GameObject as GameObject
from mathutils import Vector
from uplogic.audio import AudioSystem
from uplogic.audio.audiosystem import AudioCache
from uplogic.audio import get_audio_system
from uplogic.events import schedule_callback
from uplogic.utils.math import interpolate
from uplogic import console
from uplogic.utils import DELTA_TIME
import bpy
import aud


class ULReverb():
    """Reverb sound added by `Sound3D` on demand.
    """

    volume: float

    def __init__(
        self,
        parent,
        sound,
        handle
    ):
        self.volume = 0
        self.parent = parent
        self.handle = handle
        self.aud_system = parent.aud_system
        self.samples = []
        schedule_callback(self.add_sample, 1/60, sound)

    def add_sample(self, sound):
        handle = self.handle
        sample = self.aud_system.device.play(sound)
        self.samples.append(sample)
        sample.loop_count = handle.loop_count
        sample.position = handle.position - (.0001 * len(self.samples))
        sample.relative = handle.relative
        sample.location = handle.location
        sample.velocity = handle.velocity
        sample.attenuation = handle.attenuation
        ori = self.parent.speaker.worldOrientation.to_quaternion()
        ori.negate()
        sample.orientation = ori
        sample.pitch = handle.pitch
        sample.volume = 0
        sample.distance_reference = handle.distance_reference
        sample.distance_maximum = handle.distance_maximum
        sample.cone_angle_inner = handle.cone_angle_inner
        sample.cone_angle_outer = handle.cone_angle_outer
        sample.cone_volume_outer = handle.cone_volume_outer
        if len(self.samples) < 30:
            schedule_callback(self.add_sample, 1/60, sound)

    def update(self):
        sample_count = self.aud_system.bounces
        use_reverb = (
            self.aud_system.reverb
        )
        handle = self.handle
        if not use_reverb or sample_count == 0:
            if self.volume < .001:
                return
            else:
                self.volume = interpolate(self.volume, 0, .1)
        else:
            parent = self.parent
            target_vol = (
                parent.volume / 10 if
                parent.occluded else
                parent.volume / 2
            )
            self.volume = interpolate(self.volume, target_vol, .1)
        for idx, sample in enumerate(self.samples):
            if not sample.status:
                sample.stop()
                continue
            if idx > sample_count:
                sample.volume = 0
                continue
            mult = idx/sample_count
            loc = handle.location
            lloc = self.aud_system.device.listener_location
            loc = (loc[0]-lloc[0], loc[1]-lloc[1], loc[2]-lloc[2])
            sample.location = (
                -(loc[0]-lloc[0]),
                -(loc[1]-lloc[1]),
                -(loc[2]-lloc[2])
            )
            sample.velocity = handle.velocity
            sample.attenuation = handle.attenuation
            ori = self.parent.speaker.worldOrientation.to_quaternion()
            ori.negate()
            sample.orientation = ori
            sample.distance_maximum = handle.distance_maximum
            sample.cone_angle_inner = handle.cone_angle_inner
            sample.pitch = handle.pitch
            sample.volume = (1-(handle.volume * (mult**2)))*.5 * self.volume * self.aud_system.volume
            sample.cone_volume_outer = handle.cone_volume_outer


def dummy():
    pass


class ULSound():
    """Base class for 2D and 3D Sounds"""

    sound = None
    """Internal `aud.Sound` instance."""
    finished: bool
    """True if the sound has played to its end."""
    pitch: float
    """Pitch (Frequency Shift)."""
    volume: float
    """Volume (Amplitude)."""
    aud_system: AudioSystem
    """Audio System this sound is playing on."""

    @property
    def position(self):
        '''Progression of the soundfile in seconds.'''
        if self.sound:
            return self.sound.position

    @position.setter
    def position(self, val):
        if self.sound:
            self.sound.position = val

    def play(self):
        '''Start playback of this sound.'''
        self.sound.resume()

    def stop(self):
        '''Stop and remove this sound.'''
        self.on_finish = dummy
        self.sound.stop()

    def pause(self):
        '''Stop playback of this sound but keep it.'''
        self.sound.pause()

    def resume(self):
        '''Restart playback of this sound from the position it was paused at.'''
        self.sound.resume()

    def cache(self):
        self.aud_system.cache(self)

    def uncache(self):
        self.aud_system.uncache(self)

    def _get_soundpath(self, file):
        soundpath = file
        if not isinstance(soundpath, bpy.types.Sound):
            soundpath = bpy.data.sounds.get(file, None)
        if soundpath:
            soundpath = soundpath.filepath
        else:
            soundpath = logic.expandPath(file)
        if not isfile(soundpath):
            console.error(f'Soundfile {soundpath} could not be loaded!')
            return None
        return soundpath

    @property
    def keep(self):
        return self.sound.keep
    
    @keep.setter
    def keep(self, val):
        self.sound.keep = val

    def on_finish(self):
        '''Standart callback to be called when the sound finishes or is stopped.'''
        pass


class Sound2D(ULSound):
    '''
    Non-spacial sound, e.g. Music or Voice-Overs.\n
    This class allows for modification of pitch and volume while playing.

    :param file: Path to the sound file.
    :param volume: Initial volume.
    :param pitch: Initial pitch.
    :param loop_count: Plays the sound this many times (0 for once, -1 for endless).
    :param lowpass: Play this effect with a lowpass filter applied.
    :param ignore_timescale: Play the sound using `Sound2D.pitch`, regardless of the current timescale.
    :param aud_sys: Audiosystem to play this sound on.
    '''

    _deprecated = False

    def __init__(
        self,
        file: str = '',
        volume: float = 1,
        pitch: float = 1,
        loop_count: int = 0,
        lowpass = False,
        ignore_timescale = True,
        mono: bool = False,
        aud_sys: str = 'default'
    ):
        if self._deprecated:
            console.warning('Warning: ULSound2D class will be renamed to "Sound2D" in future releases!')
        self.file = file
        self._volume = 1
        self.finished = False
        if not (file):
            return
        self.aud_system = get_audio_system(aud_sys)
        self.ignore_timescale = ignore_timescale
        soundpath = self._get_soundpath(file)
        if soundpath is None:
            return
        sound = self.soundfile = AudioCache.get(self.file, aud.Sound(soundpath))
        lowpass = self.aud_system.lowpass or lowpass
        if lowpass:
            sound = self.soundfile = sound.lowpass(lowpass, .5)
        if mono:
            sound = sound.rechannel(1)
        device = self.aud_system.device
        self.sound = handle = device.play(sound)
        self.sound.pause()
        handle.volume = 0
        handle.relative = True
        handle.loop_count = loop_count
        self.aud_system.add(self)
        self.volume = volume
        self.pitch = pitch
        self._lowpass = False
        self.lowpass = self.aud_system.lowpass

    @property
    def panning(self):
        """Pan a mono sound from left (-1) to right (1).

        Note that this only works for mono sounds."""
        return self.sound.location[0]

    @panning.setter
    def panning(self, val):
        panning = list(self.sound.location)
        panning[0] = val
        self.sound.location = panning

    @property
    def volume(self):
        '''Playback amplitude.'''
        return self._volume

    @volume.setter
    def volume(self, val):
        if self.sound and self.sound.status:
            self.sound.volume = val * self.aud_system.volume
        self._volume = val

    @property
    def pitch(self):
        '''Playback frequency shift.'''
        ts = 1 if self.ignore_timescale else logic.getTimeScale()
        if self.sound and self.sound.status:
            return self.sound.pitch / ts

    @pitch.setter
    def pitch(self, val):
        ts = 1 if self.ignore_timescale else logic.getTimeScale()
        if self.sound and self.sound.status:
            self.sound.pitch = val * ts

    @property
    def lowpass(self):
        '''Frequency cutoff as a factor of 20.000.'''
        return self._lowpass

    @lowpass.setter
    def lowpass(self, val):
        if abs(self._lowpass - val) < 10:
            return
        self._lowpass = val
        sound = self.soundfile
        if val:
            sound = sound.lowpass(val, .5)
        sound = self.aud_system.device.play(sound)
        sound.loop_count = self.sound.loop_count
        sound.position = self.sound.position + DELTA_TIME()
        sound.volume = self.sound.volume
        sound.pitch = self.sound.pitch
        schedule_callback(self.sound.stop)
        self.sound = sound

    def update(self):
        '''This function is called each frame and updates the attributes of the sound according to the scene.'''
        if self.volume == 0:
            return
        handle = self.sound
        if not handle.status:
            self.finished = True
            self.on_finish()
            self.aud_system.remove(self)
            return


class ULSound2D(Sound2D):
    '''[DEPRECATED] Use `uplogic.audio.Sound2D` instead

    Non-spacial sound, e.g. Music or Voice-Overs.\n
    This class allows for modification of pitch and volume while playing.

    :param file: Path to the sound file.
    :param volume: Initial volume.
    :param pitch: Initial pitch.
    :param loop_count: Plays the sound this many times (0 for once, -1 for endless).
    :param aud_sys: Audiosystem to play this sound on.
    '''

    _deprecated = True


class Sample2D(Sound2D):
    '''Non-spacial sample, e.g. Music or Voice-Overs.\n
    This class allows for modification of pitch and volume while playing.
    The played audio file can be limited to a start and end time.

    :param file: Path to the sound file.
    :param sample: Tuple containing the "start" and "end" timestamp.
    :param volume: Initial volume.
    :param pitch: Initial pitch.
    :param loop_count: Plays the sound this many times (0 for once, -1 for endless).
    :param lowpass: Play this effect with a lowpass filter applied.
    :param ignore_timescale: Play the sound using `Sample2D.pitch`, regardless of the current timescale.
    :param aud_sys: Audiosystem to play this sound on.
    '''

    _deprecated = False

    def __init__(
        self,
        file: str = '',
        sample: tuple = (0, 0),
        volume: float = 1,
        pitch: float = 1,
        loop_count: int = 0,
        lowpass = False,
        ignore_timescale = False,
        mono=False,
        aud_sys: str = 'default'
    ):
        self.file = file
        self.soundfile = None
        self._volume = 1
        self.finished = False
        if not (file):
            return
        self.aud_system = get_audio_system(aud_sys)
        self.ignore_timescale = ignore_timescale
        soundpath = file
        if not isinstance(soundpath, bpy.types.Sound):
            soundpath = bpy.data.sounds.get(file, None)
        if soundpath:
            soundpath = soundpath.filepath
        else:
            soundpath = logic.expandPath(file)
        soundpath = self._get_soundpath(file)
        if soundpath is None:
            return
        sound = self.soundfile = AudioCache.get(self.file, aud.Sound(soundpath))
        if sample[1]:
            sound = sound.limit(sample[0], sample[1])
        if mono:
            sound = sound.rechannel(1)
        lowpass = self.aud_system.lowpass or lowpass
        if lowpass:
            sound = self.soundfile = sound.lowpass(lowpass, .5)
        device = self.aud_system.device
        self.sound = handle = device.play(sound)
        handle.volume = 0

        handle.relative = True
        handle.loop_count = loop_count
        self.aud_system.add(self)
        self.volume = volume
        self.pitch = pitch
        self._lowpass = False
        self.lowpass = self.aud_system.lowpass


class Sound3D(ULSound):
    '''Spacial sound, e.g. World Effects or Voices.\n
    
    :param speaker: Play the sound at a `Vector` or use a `KX_GameObject`.
    :param file: Path to the sound file.
    :param occlusion: Muffle sounds behind walls (can be bad for performance).
    :param transition_speed: Fading speed from regular to muffled.
    :param cutoff_frequency: Cutoff for muffled version as a factor of 20.000.
    :param loop_count: The amount of times the sound should be played. -1 is looped.
    :param pitch: Initial pitch.
    :param volume: Initial volume.
    :param reverb: Use conditional reverberation (performance intense).
    :param attenuation: Distance fade factor.
    :param distance_ref: Distance at which the sound is audible at 100% volume.
    :param cone_angle: Cone spread for directional sounds. Cone is aligned to the -Z axis.
    :param cone_outer_volume: Volume outside of the cone.
    :param ignore_timescale: Play the sound using `Sound3D.pitch`, regardless of the current timescale.
    :param aud_sys: Audiosystem to play this sound on.
    '''

    _deprecated = False

    @property
    def position(self):
        if self.handles:
            return self.handles[1][0].position

    @position.setter
    def position(self, val):
        for sound in self.handles[1]:
            sound.position = val

    def __init__(
        self,
        speaker: GameObject or Vector = None,
        file: str = '',
        occlusion: bool = False,
        transition_speed: float = .1,
        cutoff_frequency: float = .1,
        loop_count: int = 0,
        pitch: float = 1,
        volume: float = 1,
        reverb=False,
        attenuation: float = 1,
        distance_ref: float = 1,
        cone_angle: list[float] = [360, 360],
        cone_outer_volume: float = 0,
        ignore_timescale: bool = False,
        aud_sys: str = 'default'
    ):
        if self._deprecated:
            console.warning(f'"{self.__class__.__name__}" class will be renamed to "Sound3D" in future releases!')
        self._is_vector = isinstance(speaker, Vector)
        self.file = file
        self.soundfile = None
        self.finished = False
        if not (file and speaker):
            return
        self._clear_sound = 0 if occlusion else 1
        self._sustained = 1
        self.occluded = False
        self.sounds = []
        self.reverb_samples = None
        self.aud_system = get_audio_system(aud_sys)
        self.speaker = speaker
        self.reverb = reverb
        self.occlusion = occlusion
        self.volume = volume
        self.pitch = pitch
        self.cone_outer_volume = cone_outer_volume
        master_volume = self.aud_system.volume
        self.transition = transition_speed
        self.ignore_timescale = ignore_timescale
        soundpath = self._get_soundpath(file)
        if soundpath is None:
            return
        sound = self.soundfile = AudioCache.get(self.file, aud.Sound(soundpath).rechannel(1))
        device = self.aud_system.device
        handle = self.sound = device.play(sound)
        handle.volume = 0
        if occlusion:
            soundlow = aud.Sound.lowpass(sound, 4400 * cutoff_frequency, .5).rechannel(1)
            handlelow = device.play(soundlow)
            handlelow.volume = 0
            self.handles = [speaker, [handle, handlelow]]
        else:
            self.handles = [speaker, [handle]]
        for handle in self.handles[1]:
            handle.relative = False
            handle.location = speaker if self._is_vector else speaker.worldPosition
            if not self._is_vector and speaker.mass:
                handle.velocity = getattr(
                    speaker,
                    'worldLinearVelocity',
                    Vector((0, 0, 0))
                ) if speaker.blenderObject.game.physics_type != 'NO_COLLISION' else Vector((0, 0, 0))
            handle.attenuation = attenuation
            if not self._is_vector:
                handle.orientation = speaker.worldOrientation.to_quaternion()
            handle.pitch = pitch
            handle.volume = volume * master_volume
            handle.distance_reference = distance_ref
            handle.distance_maximum = 1000
            handle.cone_angle_inner = cone_angle[0]
            handle.cone_angle_outer = cone_angle[1]
            handle.loop_count = loop_count
            handle.cone_volume_outer = cone_outer_volume * volume * master_volume
        if self.reverb:
            self.reverb_samples = ULReverb(
                self,
                sound,
                self.handles[1][0]
            )
        self.aud_system.add(self)
        self.update(True)

    def play(self):
        '''Start playback of this sound.'''
        for sound in self.sounds:
            sound.resume()

    def stop(self):
        '''Stop and remove this sound.'''
        self.on_finish = dummy
        for sound in self.sounds:
            sound.stop()

    def pause(self):
        '''Stop playback of this sound but keep it.'''
        for sound in self.sounds:
            sound.pause()

    def resume(self):
        '''Restart playback of this sound from the position it was paused at.'''
        for sound in self.sounds:
            sound.resume()

    def update(self, init=False):
        '''This function is called each frame and updates the attributes of the sound according to the scene.'''
        aud_system = self.aud_system
        speaker = self.speaker
        if not self._is_vector and (not speaker or speaker.invalid):
            self.finished = True
            self.on_finish()
            aud_system.remove(self)
            return
        location = speaker if self._is_vector else speaker.worldPosition
        for i, handle in enumerate(self.handles[1]):
            if not handle.status:
                self.finished = True
                self.on_finish()
                aud_system.remove(self)
                return
            handle.pitch = self.pitch * (1 if self.ignore_timescale else logic.getTimeScale())
            handle.location = location
            if not self._is_vector:
                handle.orientation = (
                    speaker
                    .worldOrientation
                    .to_quaternion()
                )
                if 'volume' in dir(self.speaker.blenderObject.data):
                    handle.velocity = Vector((0, 0, 0))
                elif speaker.mass:
                    handle.velocity = getattr(speaker, 'worldLinearVelocity', Vector((0, 0, 0)))
            if self.occlusion and handle.status:
                transition = 1 if init else self.transition
                cam = self.aud_system.listener
                occluder, point, normal = cam.rayCast(
                    location,
                    cam.worldPosition,
                    xray=False
                )
                occluded = self.occluded = False
                penetration = 1
                occ_count = 0
                while occluder and occ_count < 5:
                    if occluder is speaker:
                        break
                    sound_occluder = occluder.blenderObject.get(
                        'sound_occluder',
                        True
                    )
                    if sound_occluder:
                        occluded = self.occluded = True
                        block = occluder.blenderObject.get(
                            'sound_blocking',
                            .1
                        )
                        if penetration > 0:
                            penetration -= block
                        else:
                            penetration = 0
                    occluder, point, normal = occluder.rayCast(
                        location,
                        point,
                        xray=False
                    )
                    occ_count += 1
                cs = self._clear_sound
                if occluded and cs > 0:
                    self._clear_sound -= transition
                elif not occluded and cs < 1:
                    self._clear_sound += transition
                if self._clear_sound < 0:
                    self._clear_sound = 0
                sustained = self._sustained
                if sustained > penetration:
                    self._sustained -= transition / 10
                elif sustained < penetration:
                    self._sustained += transition / 10
                mult = (
                    cs * sustained
                    if not i
                    else (1 - cs) * sustained
                )
                master_volume = self.aud_system.volume
                try:
                    handle.volume = self.volume * mult * master_volume
                
                    handle.cone_volume_outer = (
                        self.cone_outer_volume *
                        self.volume *
                        mult *
                        master_volume
                    )
                except Exception:
                    pass
            elif handle.status:
                master_volume = self.aud_system.volume
                handle.volume = self.volume * master_volume
                handle.cone_volume_outer = (
                    self.cone_outer_volume *
                    self.volume *
                    master_volume
                )

        if self.reverb_samples:
            self.reverb_samples.update()

    def pause(self):
        self.on_finish = dummy
        for sound in self.handles[1]:
            sound.pause()

    def resume(self):
        self.on_finish = dummy
        for sound in self.handles[1]:
            sound.resume()

    def stop(self):
        self.on_finish = dummy
        for sound in self.handles[1]:
            sound.stop()


class ULSound3D(Sound3D):
    _deprecated = True


class Sample3D(Sound3D):
    '''Spacial sound, e.g. World Effects or Voices.\n
    
    :param speaker: Play the sound at a `Vector` or use a `KX_GameObject`.
    :param file: Path to the sound file.
    :param sample: Tuple containing the "start" and "end" timestamp.
    :param occlusion: Muffle sounds behind walls (can be bad for performance).
    :param transition_speed: Fading speed from regular to muffled.
    :param cutoff_frequency: Cutoff for muffled version as a factor of 20.000.
    :param loop_count: The amount of times the sound should be played. -1 is looped.
    :param pitch: Initial pitch.
    :param volume: Initial volume.
    :param reverb: Use conditional reverberation (performance intense).
    :param attenuation: Distance fade factor.
    :param distance_ref: Distance at which the sound is audible at 100% volume.
    :param cone_angle: Cone spread for directional sounds. Cone is aligned to the -Z axis.
    :param cone_outer_volume: Volume outside of the cone.
    :param ignore_timescale: Play the sound using `Sample3D.pitch`, regardless of the current timescale.
    :param aud_sys: Audiosystem to play this sound on.
    '''

    _deprecated = False

    def __init__(
        self,
        speaker: GameObject or Vector = None,
        file: str = '',
        sample: tuple = (0, 0),
        occlusion: bool = False,
        transition_speed: float = .1,
        cutoff_frequency: float = .1,
        loop_count: int = 0,
        pitch: float = 1,
        volume: float = 1,
        reverb=False,
        attenuation: float = 1,
        distance_ref: float = 1,
        cone_angle: list[float] = [360, 360],
        cone_outer_volume: float = 0,
        ignore_timescale: bool = False,
        aud_sys: str = 'default'
    ):
        self._is_vector = isinstance(speaker, Vector)
        self.file = file
        self.soundfile = None
        self.finished = False
        if not (file and speaker):
            return
        self._clear_sound = 1
        self._sustained = 1
        self.occluded = False
        self.sounds = []
        self.reverb_samples = None
        self.aud_system = get_audio_system(aud_sys)
        self.speaker = speaker
        self.reverb = reverb
        self.occlusion = occlusion
        self.volume = volume
        self.pitch = pitch
        self.cone_outer_volume = cone_outer_volume
        master_volume = self.aud_system.volume
        self.transition = transition_speed
        self.ignore_timescale = ignore_timescale
        soundpath = self._get_soundpath(file)
        if soundpath is None:
            return
        sound = self.soundfile = AudioCache.get(self.file, aud.Sound(soundpath).rechannel(1))
        device = self.aud_system.device
        if sample[1]:
            sound = sound.limit(sample[0], sample[1])
        handle = device.play(sound)
        handle.volume = 0
        if occlusion:
            soundlow = aud.Sound.lowpass(sound, 4400 * cutoff_frequency, .5).rechannel(1)
            handlelow = device.play(soundlow)
            handlelow.volume = 0
            self.handles = [speaker, [handle, handlelow]]
        else:
            self.handles = [speaker, [handle]]
        for handle in self.handles[1]:
            handle.relative = False
            handle.location = speaker if self._is_vector else speaker.worldPosition
            if not self._is_vector and speaker.mass:
                handle.velocity = getattr(
                    speaker,
                    'worldLinearVelocity',
                    Vector((0, 0, 0))
                ) if speaker.blenderObject.game.physics_type != 'NO_COLLISION' else Vector((0, 0, 0))
            handle.attenuation = attenuation
            if not self._is_vector:
                handle.orientation = speaker.worldOrientation.to_quaternion()
            handle.pitch = pitch
            handle.volume = volume * master_volume
            handle.distance_reference = distance_ref
            handle.distance_maximum = 1000
            handle.cone_angle_inner = cone_angle[0]
            handle.cone_angle_outer = cone_angle[1]
            handle.loop_count = loop_count
            handle.cone_volume_outer = cone_outer_volume * volume * master_volume
        if self.reverb:
            self.reverb_samples = ULReverb(
                self,
                sound,
                self.handles[1][0]
            )
        self.aud_system.add(self)
        self.update(True)


class Speaker2D(Sound2D):
    '''Start a speaker object using its properties.\n
    
    :param speaker: `KX_GameObject` of speaker type.
    :param loop_count: The amount of times the sound should be played. -1 is looped.
    :param lowpass: Play this effect with a lowpass filter applied.
    :param ignore_timescale: Play the sound using `Speaker2D.pitch`, regardless of the current timescale.
    :param aud_sys: Audiosystem to play this sound on.
    '''

    _deprecated = False

    def __init__(
        self,
        speaker: GameObject,
        loop_count: int = 0,
        lowpass=False,
        ignore_timescale: bool = False,
        aud_sys: str = 'default'
    ):
        if self._deprecated:
            console.warning('Warning: ULSpeaker2D class will be renamed to "Speaker2D" in future releases!')
        speaker_data = speaker.blenderObject.data
        # ULSound2D()
        super().__init__(
            speaker_data.sound.filepath,
            speaker_data.volume,
            speaker_data.pitch,
            loop_count,
            lowpass,
            ignore_timescale,
            aud_sys
        )


class ULSpeaker2D(Speaker2D):
    _deprecated = True


class Speaker3D(Sound3D):
    '''Start a speaker object using its properties.\n
    
    :param speaker: `KX_GameObject` of speaker type.
    :param occlusion: Muffle sounds behind walls (can be bad for performance).
    :param transition_speed: Fading speed from regular to muffled.
    :param cutoff_frequency: Cutoff for muffled version as a factor of 20.000.
    :param loop_count: The amount of times the sound should be played. -1 is looped.
    :param reverb: Use conditional reverberation (performance intense).
    :param ignore_timescale: Play the sound using `Speaker3D.pitch`, regardless of the current timescale.
    :param aud_sys: Audiosystem to play this sound on.
    '''

    _deprecated = False

    def __init__(
        self,
        speaker: GameObject,
        occlusion: bool = False,
        transition_speed: float = 0.1,
        cutoff_frequency: float = 0.1,
        loop_count: int = 0,
        reverb=False,
        ignore_timescale: bool = False,
        aud_sys: str = 'default'
    ):
        if self._deprecated:
            console.warning('Warning: ULSpeaker3D class will be renamed to "Speaker3D" in future releases!')
        speaker_data = speaker.blenderObject.data
        super().__init__(
            speaker,
            speaker_data.sound.filepath,
            occlusion,
            transition_speed,
            cutoff_frequency,
            loop_count,
            speaker_data.pitch,
            speaker_data.volume,
            reverb,
            speaker_data.attenuation,
            speaker_data.distance_reference,
            [speaker_data.cone_angle_inner, speaker_data.cone_angle_outer],
            speaker_data.cone_volume_outer,
            ignore_timescale,
            aud_sys
        )


class ULSpeaker3D(Speaker3D):
    _deprecated = True


def play_sound_2d(
        file: str = '',
        volume: float = 1,
        pitch: float = 1,
        loop_count: int = 0,
        lowpass = False,
        ignore_timescale = True,
        mono: bool = False,
        aud_sys: str = 'default'
    ):
        sound = Sound2D(
            file=file,
            volume=volume,
            pitch=pitch,
            loop_count=loop_count,
            lowpass=lowpass,
            ignore_timescale=ignore_timescale,
            mono=mono,
            aud_sys=aud_sys
        )
        if sound.sound:
            sound.play()
        return sound


def play_sound_3d(
        speaker: GameObject or Vector = None,
        file: str = '',
        occlusion: bool = False,
        transition_speed: float = .1,
        cutoff_frequency: float = .1,
        loop_count: int = 0,
        pitch: float = 1,
        volume: float = 1,
        reverb=False,
        attenuation: float = 1,
        distance_ref: float = 1,
        cone_angle: list[float] = [360, 360],
        cone_outer_volume: float = 0,
        ignore_timescale: bool = False,
        aud_sys: str = 'default'
    ):
        sound = Sound3D(
            speaker=speaker,
            file=file,
            occlusion=occlusion,
            transition_speed=transition_speed,
            cutoff_frequency=cutoff_frequency,
            loop_count=loop_count,
            pitch=pitch,
            volume=volume,
            reverb=reverb,
            attenuation=attenuation,
            distance_ref=distance_ref,
            cone_angle=cone_angle,
            cone_outer_volume=cone_outer_volume,
            ignore_timescale=ignore_timescale,
            aud_sys=aud_sys
        )
        if sound.sound:
            sound.play()
        return sound



from uplogic.nodes import ULParameterNode


class FModGetEventAttributeNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.event = None
        self.attribute = 'position'
        self.EVT = self.add_output(self.get_evt)
        self.VALUE = self.add_output(self.get_value)

    def get_evt(self):
        return self.get_input(self.event)

    def get_value(self):
        return getattr(self.get_input(self.event), self.attribute, None)



from .shader import Filter2D


glsl = """
uniform sampler2D bgl_RenderedTexture;

// in vec4 bgl_TexCoord;
// out vec4 fragColor;
uniform float gamma;

void main()
{
    fragColor = pow(texture(bgl_RenderedTexture, bgl_TexCoord.xy), vec4(gamma));
}
"""


class Buffer(Filter2D):

    def __init__(self, gamma=1, idx: int = None) -> None:
        self.uniforms = {'gamma': float(gamma)}
        super().__init__(glsl, idx, {'gamma': self.uniforms})
        self._filter.addOffScreen(0)

    @property
    def bindcode(self):
        return self._filter.offScreen.colorBindCodes[0]

    @property
    def texture(self):
        return self._filter.offScreen.getColorTexture()


depth_glsl = """
uniform sampler2D bgl_DepthTexture;

// in vec4 bgl_TexCoord;
// out vec4 fragColor;
uniform float gamma;

void main()
{
    fragColor = pow(texture(bgl_DepthTexture, bgl_TexCoord.xy), vec4(gamma));
}
"""


class DepthBuffer(Buffer):

    def __init__(self, gamma=1, idx: int = None) -> None:
        self.uniforms = {'gamma': float(gamma)}
        super().__init__(depth_glsl, idx, {'gamma': self.uniforms})
        self._filter.addOffScreen(0)




from uplogic.nodes import ULActionNode


class SetConstraintAttributeNode(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.object = None
        self.constraint = None
        self.attribute = None
        self.value = None
        self._done: bool = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        consts = self.get_input(self.object).blenderObject.constraints
        const = consts[self.get_input(self.constraint)]
        setattr(
            const,
            self.get_input(self.attribute),
            self.get_input(self.value)
        )
        self._done = True



from uplogic.nodes import ULActionNode
from uplogic import console
from bpy.types import Material
from bge.types import KX_GameObject


class ULSetMaterial(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.slot = None
        self.mat_name = None
        self.OUT = self.add_output(self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object: KX_GameObject = self.get_input(self.game_object)
        slot: int = self.get_input(self.slot)
        material: Material = self.get_input(self.mat_name)
        bl_obj = game_object.blenderObject
        if slot > len(bl_obj.material_slots) - 1:
            console.debug('Set Material: Slot does not exist!')
            return
        bl_obj.material_slots[slot].material = material
        self._done = True



from uplogic.nodes import ULActionNode
from uplogic.utils import make_valid_name


class ToggleLogicTreePropertyNode(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.prop_name = None
        self.OUT = self.add_output(self.get_property)

    def get_property(self):
        return self._done

    def evaluate(self):
        property_name = self.get_input(self.prop_name)
        condition = self.get_condition()
        if not condition:
            return
        result = getattr(self.network.component, make_valid_name(property_name).lower(), False)
        setattr(self.network.component, make_valid_name(property_name).lower(), not result)
        
        self._done = True



from bge import logic
import bpy
import gpu
from math import floor
from os.path import isfile
from uplogic import console


class ImageHandler:

    def __init__(self, texture, fps=60, min_frame=0, max_frame=None, load_audio=False):
        self.play_mode = 'play'
        self.sound = None
        self._texture = None
        self._image = None
        self.image = None
        self._opacity = 1
        self.load_audio = load_audio
        min_frame=0
        self.fps = fps
        self._is_playing = False
        self._ref_time = 0
        self.time = 0
        self._flushed = False
        self._frame = 1
        if texture is None:
            max_frame=1000
            return
        if texture is not None and texture not in bpy.data.images and isfile(texture):
            self.image = bpy.data.images.load(texture)
        self.texture = texture

        self.min_frame = min_frame
        self.max_frame = self.image.frame_duration
        if max_frame is not None:
            self.max_frame = max_frame
        logic.getCurrentScene().pre_draw.append(self.update)

    @property
    def filepath(self):
        return self.image.filepath if self.image else ''

    @property
    def fps(self):
        return self._fps

    @fps.setter
    def fps(self, val):
        self._fps = val

    @property
    def playback_position(self):
        return self.frame / self.fps

    @playback_position.setter
    def playback_position(self, val):
        self.seek(val)

    @property
    def is_playing(self):
        return self._is_playing

    @is_playing.setter
    def is_playing(self, val):
        if not self.is_playing and val:
            self._ref_time = logic.getRealTime() - self.playback_position
            if self.sound is not None:
                self.sound.play()
                self.sound.position = self.playback_position
        elif not val and self.sound is not None:
            self.sound.pause()
        self._is_playing = val

    @property
    def texture(self):
        return self._texture

    @texture.setter
    def texture(self, val):
        if val is None:
            return
        texture = bpy.data.images.get(val, None)
        if not texture:
            texture = bpy.data.images.load(val, check_existing=True)
        self.image = texture
        self._texture = gpu.texture.from_image(texture)
        self.max_frame = self.image.frame_duration
        if self.load_audio:
            if self.sound is not None:
                self.sound.stop()
            try:
                from uplogic.audio import Sound2D
                self.sound = Sound2D(self.filepath)
                self.sound.keep = True
            except Exception:
                self.sound = None
                console.warning("Couldn't read audio from movie file.")

    @property
    def frame(self):
        return self._frame

    @frame.setter
    def frame(self, val):
        self._frame = max(self.min_frame, min(val, self.max_frame))
        self.flush()

    def play(self):
        self.is_playing = True

    def seek(self, position):
        self._ref_time = logic.getRealTime() - position
        self.time = logic.getRealTime() - self._ref_time
        self.frame = floor(self.time * self._fps)
        if self.sound is not None:
            self.sound.position = self.playback_position
            if self.is_playing:
                self.sound.play()

    def flush(self):
        if not self._flushed:
            self.image.gl_free()
            self.image.gl_load(frame=self.frame)

            self._texture = gpu.texture.from_image(self.image)
            self.image.update_tag()

    def update(self):
        if self.is_playing:
            self.time = logic.getRealTime() - self._ref_time
            self.frame = int(self.time * self._fps)
            if self.frame >= self.max_frame:
                self._finish()

    def _finish(self):
        if self.play_mode == 'loop':
            self.seek(0)
        else:
            self.is_playing = False
        self.on_finish()

    def on_finish(self):
        ...

    def free(self):
        self.image.gl_free()
        self.image.buffers_free()

    def stop(self):
        self.free()
        if self.sound is not None:
            self.sound.stop()
        self.is_playing = False


from uplogic.nodes import ULConditionNode


class ULAndNot(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.condition_a = None
        self.condition_b = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        ca = self.get_input(self.condition_a)
        cb = not self.get_input(self.condition_b)
        return ca and not cb



from uplogic.nodes import ULParameterNode


class ULListDuplicate(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.condition = None
        self.items = None
        self.OUT = self.add_output(self.get_points)

    def get_points(self):
        return self.get_input(self.items).copy()



from uplogic.nodes import ULActionNode
from uplogic import console
import bpy
import json
import os


class ULClearVariables(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = False
        self.file_name = ''
        self.path = ''
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def write_to_json(self, path):
        data = None
        if not path.endswith('.json'):
            path = os.path.join(path, f'{self.get_input(self.file_name)}.json')
        if os.path.isfile(path):
            data = {}
            f = open(path, 'w')
            json.dump(data, f, indent=2)
        else:
            console.debug('File does not exist - creating...')
            f = open(path, 'w')
            data = {}
            json.dump(data, f, indent=2)
        f.close()

    def evaluate(self):
        condition = self.get_condition()
        if not condition:
            return
        path = self.get_input(self.path)

        path = (bpy.path.abspath(path=path))
        os.makedirs(path, exist_ok=True)

        self.write_to_json(path)
        self._done = True



from uplogic.nodes import ULParameterNode
from bpy.types import Collection
from bge.logic import getCurrentScene


class ULGetCollectionObjects(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.collection = None
        self.OUT = self.add_output(self.get_objects)

    def get_objects(self):
        collection: Collection = self.get_input(self.collection)
        scene = getCurrentScene()
        return [scene.getGameObjectFromObject(o) for o in collection.objects]



from uplogic.nodes import ULActionNode
from bge.types import KX_GameObject
from bge.logic import sendMessage


class ULSendMessage(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.from_obj = None
        self.to_obj = None
        self.subject = None
        self.body = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        from_obj: KX_GameObject = self.get_input(self.from_obj)
        to_obj: KX_GameObject = self.get_input(self.to_obj)
        subject = self.get_input(self.subject)
        body = self.get_input(self.body)
        sendMessage(
            subject,
            body,
            to_obj.name if to_obj else '',
            from_obj.name if from_obj else ''
        )
        self._done = True



from uplogic.nodes import ULActionNode


class ULSetProperty(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.property_name = None
        self.property_value = None
        self.mode = 0
        self.OUT = self.add_output(self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object = self.get_input(self.game_object)
        property_name = self.get_input(self.property_name)
        property_value = self.get_input(self.property_value)
        self._done = True
        obj = game_object.blenderObject if self.mode else game_object
        obj[property_name] = property_value
        if self.mode:
            obj.color = obj.color  # force object data refresh



from bge import logic
from uplogic.nodes import ULActionNode
from uplogic.loop import get_mainloop


class ULEndGame(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None

    def evaluate(self):
        if not self.get_condition():
            return
        mainloop = get_mainloop()
        if mainloop:
            mainloop.stop()
        logic.endGame()



from mathutils import Euler
from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic import console


class ULSetBonePosition(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.armature = None
        self.bone_name = None
        self.set_translation = None
        self._eulers = Euler((0, 0, 0), "XYZ")
        self._vector = Vector((0, 0, 0))
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        armature = self.get_input(self.armature)
        bone_name = self.get_input(self.bone_name)
        set_translation = self.get_input(self.set_translation)
        # TODO cache the bone index
        bone_channel = armature.channels[bone_name]
        if set_translation is not None:
            bone_channel.location = set_translation
        else:
            console.debug('Set Bone Node: Position is None!')
        armature.update()
        self._done = True



from bge import constraints
from bge import logic
from uplogic.nodes import ULParameterNode
import bpy


class ULCharacterInfo(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.max_jumps = None
        self.cur_jump = None
        self.gravity = None
        self.walk_dir = None
        self.on_ground = None
        self._physics = None
        self.local = False
        self.MAX_JUMPS = self.add_output(self.get_max_jumps)
        self.CUR_JUMP = self.add_output(self.get_current_jump)
        self.GRAVITY = self.add_output(self.get_gravity)
        self.WALKDIR = self.add_output(self.get_walk_dir)
        self.ON_GROUND = self.add_output(self.get_on_ground)

    def get_max_jumps(self):
        return self.physics.maxJumps

    def get_current_jump(self):
        return self.physics.jumpCount

    def get_gravity(self):
        return self.physics.gravity

    def get_walk_dir(self):
        physics = self.physics
        wdir = (
            physics.walkDirection @ self.owner.worldOrientation
            if self.local else
            physics.walkDirection
        )
        return wdir * bpy.data.scenes[
            logic.getCurrentScene().name
        ].game_settings.physics_step_sub

    def get_on_ground(self):
        return self.physics.onGround

    def fetch(self):
        game_object = self.owner = self.get_input(self.game_object)
        self.physics = constraints.getCharacter(game_object)



from uplogic.nodes import ULActionNode
from uplogic.utils import rotate_to


class ULActionRotateTo(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.moving_object = None
        self.target_point = None
        self.speed = None
        self.rot_axis = 2
        self.front_axis = 0
        self.OUT = self.add_output(self._get_done)

    def _get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        moving_object = self.get_input(self.moving_object)
        target_point = self.get_input(self.target_point)
        speed = self.get_input(self.speed)
        rot_axis = self.get_input(self.rot_axis)
        front_axis = self.get_input(self.front_axis)
        rotate_to(moving_object, target_point, rot_axis, front_axis, speed)



from uplogic.nodes import ULActionNode
from uplogic.ui import Path
from math import degrees


class CreateUIPathNode(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.parent = None
        self.points = []
        self.rel_pos = False
        self.pos = (0, 0)
        self.rel_points = False
        self.angle = 0
        self.line_color = (9, 0, 0, 0)
        self.line_width = 0
        self._widget = None
        self.DONE = self.add_output(self._get_done)
        self.PATH = self.add_output(self._get_widget)

    def _get_done(self):
        return self._done

    def _get_widget(self):
        return self._widget

    def evaluate(self):
        condition = self.get_condition()
        if not condition:
            return
        ipt = self.get_input
        parent = ipt(self.parent)
        rel_pos = ipt(self.rel_pos)
        pos = ipt(self.pos)
        rel_points = ipt(self.rel_points)
        points = ipt(self.points)
        line_color = ipt(self.line_color)
        line_width = ipt(self.line_width)
        angle = degrees(ipt(self.angle))

        self._widget = Path(
            pos=pos,
            relative={'pos': rel_pos, 'points': rel_points},
            points=points,
            line_color=line_color,
            line_width=line_width,
            angle=angle
        )
        if parent:
            parent.add_widget(self._widget)
        self._done = True
        



from uplogic.nodes import ULConditionNode
from uplogic.input import mouse_tap
from uplogic.input import mouse_down
from uplogic.input import mouse_up


class ULMousePressed(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.pulse = False
        self.input_type = 1 if self.pulse else 0
        self.mouse_button_code = None
        self.OUT = self.add_output(self.get_pressed)

    def get_pressed(self):
        mouse_button = self.get_input(self.mouse_button_code)
        funcs = [mouse_tap, mouse_down, mouse_up]
        return funcs[self.input_type](mouse_button)



from bge import logic
from uplogic.events import schedule
from bge.types import KX_GameObject as GameObject
import bpy
from mathutils import Vector, Matrix
from .math import get_bitmask
from .raycasting import raycast, raycast_projectile
from .math import cycle
from .visuals import draw_cube
from .constants import RED, WHITE
from uplogic import console


class Spawn:
    def __init__(self, object: GameObject, pool) -> None:
        schedule(self.destroy, pool._lifetime)
        schedule(self._reset_physics)
        self._pool = pool
        self._object = None
        self.scene = logic.getCurrentScene()
        if object and object['spawn'] is None:
            self._object = object
            object.restorePhysics()
            object['spawn'] = self
        if pool.spawner:
            self.transform = pool.spawner.worldTransform.copy()
            self.position = pool.spawner.worldPosition.copy()
            self.orientation = pool.spawner.worldOrientation.copy()
        else:
            self.transform = Matrix()
            self.position = Vector((0, 0, 0))
            self.orientation = Matrix([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
        scene = logic.getCurrentScene()
        if self._update not in scene.pre_draw:
            scene.pre_draw.append(self._update)
        self._visualize = False
        self.start()

    def _reset_physics(self):
        object = self._object
        if not object:
            return
        object.worldScale = (1, 1, 1)
        object.worldLinearVelocity = (0, 0, 0)
        object.worldAngularVelocity = (0, 0, 0)


    @property
    def transform(self):
        return self._transform
    
    @transform.setter
    def transform(self, pos):
        self._transform = pos
        if self._object:
            self._object.worldTransform = self._transform

    @property
    def position(self):
        return self._position
    
    @position.setter
    def position(self, pos):
        self._position = pos
        if self._object:
            self._object.worldPosition = self._position

    @property
    def orientation(self):
        return self._orientation
    
    @orientation.setter
    def orientation(self, ori):
        self._orientation = ori
        if self._object:
            self._object.worldOrientation = self._orientation

    @property
    def game_object(self):
        return self._object

    @game_object.setter
    def game_object(self, val):
        console.debug("Attribute 'game_object' of 'Spawn' is read-only!")

    def start(self):
        pass

    def _update(self):
        if self._visualize:
            draw_cube(self.position, 1, centered=True)
        self.update()

    def update(self):
        pass

    def destroy(self):
        scene = logic.getCurrentScene()
        if self._update in scene.pre_draw:
            scene.pre_draw.remove(self._update)
        obj = self._object
        if obj and obj['spawn'] is self:
            obj.worldPosition = self._pool._reset_pos
            obj.worldScale = (.001, .001, .001)
            obj['spawn'] = None
            obj.suspendPhysics()
        self.stop()

    def stop(self):
        pass


class SpawnedInstance(Spawn):
    def __init__(self, object: GameObject, pool) -> None:
        super().__init__(object=object, pool=pool)
        # schedule(self.destroy, pool._lifetime)
    #     self._pool = pool
    #     self._object = None
    #     self.scene = logic.getCurrentScene()
    #     if object and object['spawn'] is None:
    #         self._object = object
    #         object.worldPosition = (0, 0, 0)
    #         object.worldScale = (1, 1, 1)
    #         object['spawn'] = self
    #         object.restorePhysics()
    #         object.worldLinearVelocity = (0, 0, 0)
    #         object.worldAngularVelocity = (0, 0, 0)
    #     if pool.spawner:
    #         self.transform = pool.spawner.worldTransform.copy()
    #         self.position = pool.spawner.worldPosition.copy()
    #         self.orientation = pool.spawner.worldOrientation.copy()
    #     else:
    #         self.transform = Matrix()
    #         self.position = Vector((0, 0, 0))
    #         self.orientation = Matrix([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
    #     scene = logic.getCurrentScene()
    #     if self._update not in scene.pre_draw:
    #         scene.pre_draw.append(self._update)
    #     self._visualize = False
    #     self.start()

    def _reset_physics(self):
        super()._reset_physics()
        for c in self._object.components:
            tree = getattr(c, 'logictree', None)
            if tree:
                tree.network._initialized = False

    @property
    def transform(self):
        return self._object.worldTransform if self._object else self._transform
    
    @transform.setter
    def transform(self, pos):
        self._transform = pos
        if self._object:
            self._object.worldTransform = self._transform

    @property
    def position(self):
        return self._object.worldPosition if self._object else self._position
    
    @position.setter
    def position(self, pos):
        self._position = pos
        if self._object:
            self._object.worldPosition = self._position

    @property
    def orientation(self):
        return self._object.worldOrientation if self._object else self._orientation
    
    @orientation.setter
    def orientation(self, ori):
        self._orientation = ori
        if self._object:
            self._object.worldOrientation = self._orientation

    @property
    def game_object(self):
        return self._object

    @game_object.setter
    def game_object(self, val):
        console.debug("Attribute 'game_object' of 'Spawn' is read-only!")

    def start(self):
        pass

    def _update(self):
        if self._visualize:
            draw_cube(self.position, 1, color=WHITE if self._object else RED, centered=True)
        self.update()

    def update(self):
        pass

    def destroy(self):
        scene = logic.getCurrentScene()
        if self._update in scene.pre_draw:
            scene.pre_draw.remove(self._update)
        obj = self._object
        if obj and obj['spawn'] is self:
            obj.worldPosition = self._pool._reset_pos
            obj.worldScale = (.001, .001, .001)
            obj['spawn'] = None
            obj.suspendPhysics()
        self.stop()

    def stop(self):
        pass


class SimpleBullet(Spawn):
    speed = 20
    
    def update(self):
        target = self.position + Vector((0, self.speed, 0)) @ self.transform.inverted()
        dat = raycast(
            self.game_object if self.game_object else self.scene.active_camera,
            self.position,
            target,
            mask=self._pool.raycast_mask
        )
        
        if not dat.obj:
            self.position = target
        else:
            self.hit(dat)
            self.destroy()
    
    def hit(self, data):
        pass


class PhysicsBullet(Spawn):
    speed = 30
    
    def start(self):
        self.target = Vector((0, 1, 0)) @ self.transform.inverted()
    
    def update(self):
        dat = raycast_projectile(
            self.game_object if self.game_object else self.scene.active_camera,
            self.position,
            self.target,
            power=self.speed * 10,
            distance=self.speed,
            visualize=self._visualize,
            local=True,
            mask=self._pool.raycast_mask
        )
        if self.game_object:
            self.game_object.alignAxisToVect(self.target, 1, 1.0)
        
        if not dat.obj:
            self.position = dat.points[-1]
            self.target = dat.points[-1] - dat.points[-2]
        else:
            self.hit(dat)
            self.destroy()
    
    def hit(self, data):
        pass


class SpawnPool:
    _spawn_idx = 0

    def __init__(
        self,
        object_name: str,
        amount: int = 10,
        lifetime: float = 5,
        spawner = None,
        spawn = Spawn,
        raycast_mask=get_bitmask(all=True),
        inactive_pos: Vector = Vector((0, 0, -100)),
        visualize: bool = False
    ):
        self.spawn_cls = spawn
        # self._spawn_idx = 0
        self.spawner = spawner
        self.raycast_mask = raycast_mask
        self._amount = amount # clamp(amount, 0, 10)
        self._lifetime = lifetime
        
        self.visualize = visualize
        if object_name:
            bobj = bpy.data.objects[object_name]
            self._spawn_name = f'Spawned{bobj.name}'
            self.scene = logic.getCurrentScene()
            self._reset_pos = inactive_pos
            for x in range(self._amount):
                if self.scene.objects.get(f'{self._spawn_name}{x}'):
                    continue
                if spawn is SpawnedInstance:
                    bspawn = bobj.copy()
                    bspawn.name = f'{self._spawn_name}{x}'
                else:
                    bspawn = bpy.data.objects.new(f'{self._spawn_name}{x}', bobj.data)
                    bspawn.game.physics_type = bobj.game.physics_type
                bpy.context.collection.objects.link(bspawn)
                bspawn.location = Vector(inactive_pos)
                bspawn.scale = Vector((.001, .001, .001))
                gobj = self.scene.convertBlenderObject(bspawn)
                gobj['spawn'] = None
                gobj.suspendPhysics()
        else:
            self._spawn_name = None

    def spawn(self):
        spawn_obj = self.scene.objects[f'{self._spawn_name}{self._spawn_idx}'] if self._spawn_name else None
        spawn = self.spawn_cls(spawn_obj, self)
        spawn._visualize = self.visualize
        self.__class__._spawn_idx = cycle(self._spawn_idx + 1, 0, self._amount - 1)
        return spawn




from .widget import Widget
import gpu
import bge, bpy
from math import ceil
from uplogic.utils.math import rotate2d
from gpu_extras.batch import batch_for_shader
from mathutils import Vector


class Circle(Widget):

    vertex_shader = """
    in vec2 texCoord;
    in vec2 pos;
    out vec2 uv;

    uniform mat4 ModelViewProjectionMatrix;

    void main() {
        uv = texCoord;
        gl_Position = ModelViewProjectionMatrix * vec4(pos.xy, 0.0, 1.0);
    }
    """

    fragment_shader = """
    in vec2 uv;
    out vec4 fragColor;

    uniform sampler2D image;
    uniform float alpha = 1.0;

    void main() {
        vec4 color = mix(vec4(0.0), texture(image, uv), alpha);
        fragColor = pow(color, vec4(.5));
    }
    """

    def __init__(self, pos=[0, 0], radius=100, width=20, relative={}, halign='left', valign='bottom'):
        self._radius = radius
        super().__init__(pos, size=(radius, radius), relative=relative, halign=halign, valign=valign, angle=0)
        self._opacity = 1
        self.width = width
        self.start()

    @property
    def width(self):
        """Horizontal size of this widget in either pixels or factor relative to its parent."""
        return self._width

    @width.setter
    def width(self, val):
        self._width = val
        if self.parent and self.show:
            self._rebuild = True

    @property
    def radius(self):
        """Horizontal size of this widget in either pixels or factor relative to its parent."""
        return self._radius

    @radius.setter
    def radius(self, val):
        self._radius = val
        self.size = (val, val)
        if self.parent and self.show:
            self._rebuild = True

    def _build_shader(self):
        pos = self._draw_pos
        size = self._draw_size
        _resolution = 64

        points = [Vector((pos[0], pos[1] + self.radius))]
        for x in range(_resolution):
            points.append(
                rotate2d(points[-1], self.pos, 360 / _resolution)
            )
        vertices = self._vertices = points
        self._shader = gpu.shader.from_builtin('UNIFORM_COLOR')
        self._batch_line = batch_for_shader(self._shader, 'LINE_STRIP', {"pos": vertices})

    def draw(self):
        self._setup_draw()
        gpu.state.line_width_set(self.width)
        # col = self.line_color.copy()
        col = Vector((0, 1, 0, 1))
        col[3] *= self.opacity
        self._shader.uniform_float("color", col)
        self._batch_line.draw(self._shader)
        super().draw()


from uplogic.nodes import ULParameterNode
from os import path, pardir
from ...console import error
import bpy


class JoinPathNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.items = []
        self.PATH = self.add_output(self._get_path)

    def _get_path(self):
        get_ipt = self.get_input
        items = get_ipt(self.items)
        items = [get_ipt(item) for item in self.items]
        return path.join(*items)




from uplogic.nodes import ULParameterNode
from uplogic.utils import world_to_screen


class ULScreenPosition(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.camera = None
        self.pos = None
        self.OUT = self.add_output(self.get_pos)

    def get_pos(self):
        game_object = self.get_input(self.game_object)
        return world_to_screen(game_object)




from uplogic.nodes import ULActionNode


class ULApplyTransform(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.game_object = None
        self.vec1 = None
        self.vec2 = None
        self.vec3 = None
        self.local = False
        self.mode = 0
        self._actions = [
            self.apply_movement,
            self.apply_rotation,
            self.apply_force,
            self.apply_torque,
            self.apply_impulse,
        ]
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def apply_movement(self, obj, local):
        obj.applyMovement(self.get_input(self.vec1), local)

    def apply_rotation(self, obj, local):
        obj.applyRotation(self.get_input(self.vec3), local)

    def apply_force(self, obj, local):
        obj.applyForce(self.get_input(self.vec1), local)

    def apply_torque(self, obj, local):
        obj.applyTorque(self.get_input(self.vec1), local)

    def apply_impulse(self, obj, local):
        obj.applyImpulse(self.get_input(self.vec1), self.get_input(self.vec2), local)

    def evaluate(self):
        if not self.get_condition():
            return
        game_object = self.get_input(self.game_object)
        local = self.local
        self._actions[self.mode](game_object, local)
        self._done = True



from mathutils import Vector
from uplogic.nodes import ULParameterNode


class ULMath(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.operand_a = None
        self.operand_b = None
        self.operator = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        a = self.get_input(self.operand_a)
        b = self.get_input(self.operand_b)
        return self.operator(a, b)

    def get_vec_calc(self, vec, num):
        if len(vec) == 4:
            return Vector(
                (
                    self.operator(vec.x, num),
                    self.operator(vec.y, num),
                    self.operator(vec.z, num),
                    self.operator(vec.w, num)
                )
            )
        else:
            return Vector(
                (
                    self.operator(vec.x, num),
                    self.operator(vec.y, num),
                    self.operator(vec.z, num)
                )
            )

    def get_vec_vec_calc(self, vec, vec2):
        if len(vec) == 4 and len(vec2) == 4:
            return Vector(
                (
                    self.operator(vec.x, vec2.x),
                    self.operator(vec.y, vec2.y),
                    self.operator(vec.z, vec2.z),
                    self.operator(vec.w, vec2.w)
                )
            )
        else:
            return Vector(
                (
                    self.operator(vec.x, vec2.x),
                    self.operator(vec.y, vec2.y),
                    self.operator(vec.z, vec2.z)
                )
            )



from uplogic.nodes import ULParameterNode


class ULListIndex(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.items = None
        self.index = None
        self.OUT = self.add_output(self.get_val)

    def get_val(self):
        list_d = self.get_input(self.items)
        index = self.get_input(self.index)
        if index <= len(list_d) - 1:
            return list_d[index]
        return None



from uplogic.audio import Sound3D, Sound2D, Sample3D, Sample2D
from uplogic.nodes import ULActionNode
from mathutils import Vector
import bpy


class StartSoundNode(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = False
        self.sound = None
        self.start_time = 0.0
        self.end_time = 1.0
        # self.position = Vector((0, 0, 0))
        self.occlusion = False
        self.reverb = False
        self.transition = .1
        self.cutoff = .1
        self.speaker = None
        self.device = None
        self.loop_count = None
        self.pitch = 1.0
        self.speed = 1.0
        self.volume = 1.0
        self.lowpass = 1.0
        self.attenuation = 1.0
        self.distance_ref = 1.0
        self.cone_angle = Vector((360, 360))
        self.cone_outer_volume = 1.0
        self.ignore_timescale = True
        self.on_finish = False
        self.mode = 0
        self.update_running = False
        self._clear_sound = 1
        self._sustained = 1
        self._handle = None
        self.DONE = self.add_output(self.get_done)
        self.ON_FINISH = self.add_output(self.get_on_finish)
        self.HANDLE = self.add_output(self.get_handle)
        self._start = [
            self.start_2D,
            self.start_2D_sample,
            self.start_3D,
            self.start_3D_sample
        ]

    @property
    def soundfile(self):
        sound = self.get_input(self.sound)
        return sound if isinstance(sound, bpy.types.Sound) else sound

    def get_handle(self):
        return self._handle

    def get_on_finish(self):
        if not self._handle:
            return False
        if self._handle.finished:
            return True
        return False

    def get_done(self):
        return self._done
    
    def reset(self):
        super().reset()
        if self._handle and self._handle.finished:
            self._handle = None

    def evaluate(self):
        self.on_finish = False
        volume = self.get_input(self.volume)
        pitch = self.get_input(self.pitch)
        if self._handle and self.update_running:
            self._handle.volume = volume
            self._handle.pitch = pitch
            if hasattr(self._handle, 'lowpass'):
                self._handle.lowpass = self.get_input(self.lowpass) * 20000
        if not self.get_condition():
            return
        self._start[self.mode](volume, pitch)
        self._done = True

    def start_2D(self, volume, pitch):
        self._handle = Sound2D(
            file=self.get_input(self.soundfile),
            volume=volume,
            pitch=pitch,
            loop_count=self.get_input(self.loop_count),
            lowpass=self.get_input(self.lowpass) * 20000,
            ignore_timescale=self.get_input(self.ignore_timescale),
            aud_sys='ln_audio_system'
        )
        self._handle.play()

    def start_2D_sample(self, volume, pitch):
        start = self.get_input(self.start_time)
        end = self.get_input(self.end_time)
        self._handle = Sample2D(
            file=self.get_input(self.soundfile),
            sample=(start, end),
            volume=volume,
            pitch=pitch,
            loop_count=self.get_input(self.loop_count),
            lowpass=self.get_input(self.lowpass) * 20000,
            ignore_timescale=self.get_input(self.ignore_timescale),
            aud_sys='ln_audio_system'
        )
        self._handle.play()

    def start_3D(self, volume, pitch):
        self._handle = Sound3D(
            speaker=self.get_input(self.speaker),
            file=self.get_input(self.soundfile),
            occlusion=self.get_input(self.occlusion),
            transition_speed=self.get_input(self.transition),
            cutoff_frequency=self.get_input(self.cutoff),
            loop_count=self.get_input(self.loop_count),
            pitch=pitch,
            volume=volume,
            reverb=self.get_input(self.reverb),
            attenuation=self.get_input(self.attenuation),
            distance_ref=self.get_input(self.distance_ref),
            cone_angle=self.get_input(self.cone_angle),
            cone_outer_volume=self.get_input(self.cone_outer_volume),
            ignore_timescale=self.get_input(self.ignore_timescale),
            aud_sys='ln_audio_system'
        )
        self._handle.play()

    def start_3D_sample(self, volume, pitch):
        start = self.get_input(self.start_time)
        end = self.get_input(self.end_time)
        self._handle = Sample3D(
            speaker=self.get_input(self.speaker),
            file=self.get_input(self.soundfile),
            sample=(start, end),
            occlusion=self.get_input(self.occlusion),
            transition_speed=self.get_input(self.transition),
            cutoff_frequency=self.get_input(self.cutoff),
            loop_count=self.get_input(self.loop_count),
            pitch=pitch,
            volume=volume,
            reverb=self.get_input(self.reverb),
            attenuation=self.get_input(self.attenuation),
            distance_ref=self.get_input(self.distance_ref),
            cone_angle=self.get_input(self.cone_angle),
            cone_outer_volume=self.get_input(self.cone_outer_volume),
            ignore_timescale=self.get_input(self.ignore_timescale),
            aud_sys='ln_audio_system'
        )
        self._handle.play()



from uplogic.nodes import ULActionNode


class ULSlowFollow(ULActionNode):
    def __init__(self, value_type='worldPosition'):
        ULActionNode.__init__(self)
        self.value_type = str(value_type)
        self.condition = NotImplementedError
        self.game_object = None
        self.target = None
        self.factor = None
        self.OUT = self.add_output(self.get_done)

    def get_done(self):
        return self._done

    def evaluate(self):
        if not self.get_condition():
            return
        game_object = self.get_input(self.game_object)
        target = self.get_input(self.target)
        attribute = self.get_input(self.value_type)
        factor = self.get_input(self.factor)
        data = getattr(game_object, attribute)
        t_data = getattr(target, attribute)
        setattr(
            game_object,
            attribute,
            data.lerp(t_data, factor)
        )
        if attribute == 'worldScale':
            game_object.reinstancePhysicsMesh(
                game_object,
                game_object.meshes[0]
            )
        self._done = True



from uplogic.nodes import ULActionNode
from uplogic.console import write


class ULPrintValue(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        self.value = None
        self.msg_type = 'INFO'
        self.OUT = self.add_output(self.get_done)

    def evaluate(self):
        if not self.get_condition():
            return
        value = self.get_input(self.value)
        write(value, type=self.msg_type)
        self._done = True



from uplogic.nodes import ULParameterNode
import random


class ULListIndexRandom(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.condition = None
        self._item = None
        self.items = None
        self.OUT = self.add_output(self.get_val)

    def get_val(self):
        list_d = self.get_input(self.items)
        self._item = random.choice(list_d)
        return self._item



from mathutils import Vector
from ..utils.objects import GameObject
from bge.types import KX_NavMeshObject, KX_GameObject
from ..utils.visualize import draw_path


class NavPath(list[Vector]):
    """Sequence of points used for NavMesh Pathfinding."""
    # def __init__(self):
    #     self.points: list[Vector] = []
    
    # @property
    # def points(self) -> list[Vector]:
    #     return self._points

    # @points.setter
    # def points(self, val: list[Vector]):
    #     self._points = val


class NavContainer(GameObject):
    
    def __init__(self, game_object: KX_GameObject):
        super().__init__(game_object)
        self.height = 0
        self.bevel = 0.0
        self._path = NavPath()

    @property
    def next_point(self) -> Vector:
        if self._path:
            return self._path[0]

    def find_path(self, start: Vector, target: Vector, navmesh: KX_NavMeshObject):
        height = Vector((0, 0, self.height))
        points: list[Vector] = [Vector(p) + height for p in navmesh.findPath(
            start,
            target
        )]

        bevel = self.bevel
        if bevel:
            _points = self._path
            _points.clear()
            for i, p in enumerate(points):
                if i == 0 or i == len(points) - 1:
                    _points.append(p)
                    continue
                direction = (p - _points[-1]).normalized()
                next_point = points[i+1]
                next_direction = (next_point - p).normalized()
                start_circle = p - direction * bevel
                end_circle = p + next_direction * bevel

                # XXX KEEP for debugging!
                # up = direction.cross(next_direction).normalized()
                # normal1 = direction.cross(up).normalized()
                # normal2 = next_direction.cross(up).normalized()
                # draw_arrow(start_circle, start_circle - normal1)
                # draw_arrow(end_circle, end_circle - normal2)

                _points.append(start_circle)
                _points.append(end_circle)
            points = _points

        self._path = points
        return self._path

    def visualize(self, color=Vector((1, 1, 1, 1))):
        draw_path(self._path, color)

    def pop(self, idx=0):
        return self._path.pop(idx)

    def distance(self, position: Vector):
        if self._path:
            return (self.next_point - position).length
        else:
            return Vector((0, 0, 0))

    def direction(self, position: Vector):
        if self._path:
            return (self.next_point - position).normalized()
        else:
            return Vector((0, 0, 0))


class NavMesh(NavContainer):
    def __init__(self, game_obj: KX_NavMeshObject):
        self.game_object: KX_NavMeshObject = game_obj
        self._path = NavPath()

    def find_path(self, start: Vector, target: Vector):
        return super().find_path(start, target, self.game_object)



from uplogic.nodes import ULConditionNode
from bge.logic import getRealTime


class ULTimeDelay(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.condition = None
        self.delay = None
        self.triggers = []
        self.result = False
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        return self.result

    def evaluate(self):
        self.result = False
        condition = self.get_input(self.condition)
        delay = self.get_input(self.delay)
        now = getRealTime()
        if condition:
            self.triggers.append(now + delay)
        if not self.triggers:
            return
        t = self.triggers[0]
        if now >= t:
            self.result = True
            self.triggers.pop(0)
            return



from uplogic.nodes import ULParameterNode
from uplogic.utils import compute_distance


class ULDistance(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.parama = None
        self.paramb = None
        self.OUT = self.add_output(self.get_out)

    def get_out(self):
        parama = self.get_input(self.parama)
        paramb = self.get_input(self.paramb)
        return compute_distance(parama, paramb)



from bge import logic
from bge import render
from bge.types import KX_GameObject as GameObject
from bge.types import KX_PolyProxy
from mathutils import Vector
from .math import clamp
from uplogic import console


def ray_data(
    origin: Vector,
    dest: Vector,
    local: bool,
    dist: float
):
    """Get necessary data to calculate the ray.\n
    Not intended for manual use.
    """
    origin = getattr(origin, 'worldPosition', origin)
    dest = getattr(dest, 'worldPosition', dest)
    if local:
        dest = origin + dest
    d = dest - origin
    d.normalize()
    dist = dist if dist else (origin - dest).length
    dest = origin + d * dist
    return d, dist, dest


class RayCastData(tuple):
    @property
    def obj(self) -> GameObject:
        return self[0]

    @property
    def point(self) -> Vector:
        return self[1]

    @property
    def normal(self) -> Vector:
        return self[2]

    @property
    def direction(self) -> Vector:
        return self[3]

    @property
    def face(self) -> KX_PolyProxy:
        return self[4]

    @property
    def uv(self) -> Vector:
        return self[5]

# class RayCastDataPoly(RayCastData):



def raycast(
    caster: GameObject,
    origin: Vector,
    dest: Vector,
    distance: float = 0,
    prop: str = '',
    material: str = '',
    exclude: bool = None,
    xray: bool = False,
    local: bool = False,
    mask: int = 65535,
    face_data: bool = False,
    visualize: bool = False
) -> RayCastData[GameObject, Vector, Vector, Vector, KX_PolyProxy, Vector]:
    """Raycast from any point to any target

    :param caster: casting object, this object will be ignored by the ray.
    :param origin: origin point; any vector or list.
    :param dest: target point; any vector or list.
    :param distance: distance the ray will be cast
    (0 means the ray will only be cast to target).
    :param prop: look only for objects with this property.
    :param material: look only for objects with this material applied.
    :param exclude: [DEPRECATED] invert the selection for `prop` and `material`.
    :param xray: look for objects behind others.
    :param local: add the target vector to the origin.
    :param mask: Collision Mask for this ray.
    :param face_data: Return additional information about the target polygon.
    :param visualize: Show the raycast.

    :returns: (`obj`, `point`, `normal`, `direction`, `face`, `uv`)
    """
    if exclude is not None:
        from ..console import warning
        warning("raycast parameter 'exclude' is deprecated and will be removed in future versions!")
        # exclude_prop, prop = prop, ''
    origin = getattr(origin, 'worldPosition', Vector(origin)).copy()
    dest = getattr(dest, 'worldPosition', Vector(dest)).copy()
    direction, distance, dest = ray_data(origin, dest, local, distance)
    ret_dat = [None, None, None, direction, None, None]
    data = caster.rayCast(
        dest,
        objfrom=origin,
        dist=distance,
        prop=prop,
        xray=xray,
        mask=mask,
        poly=2 if face_data else 0
    )
    obj, point = data[0], data[1]
    if (material and point):
        bo = obj.blenderObject
        leftover_dist = distance - (origin - point).length
        while (
            material not in [
                slot.material.name for
                slot in
                bo.material_slots
            ]
        ) and leftover_dist > 0:
            if not xray:
                data = [None, None, None, direction, None, None]
                break
            elif point:
                old_point = point
                data = obj.rayCast(
                    dest,
                    point,
                    leftover_dist,
                    prop,
                    xray=xray,
                    mask=mask,
                    poly=2 if face_data else 0
                )
                obj, point = data[0], data[1]
                if not obj:
                    break
                bo = obj.blenderObject
                leftover_dist -= (origin - old_point).length
            else:
                data = [None, None, None, direction, None, None]
                break
    ret_dat[0] = data[0]
    ret_dat[1] = data[1]
    ret_dat[2] = data[2]
    if face_data and len(data) > 3:
        ret_dat[4] = data[3]
        ret_dat[5] = data[4]
    data = RayCastData(ret_dat)
    if visualize:
        line_dest: Vector = direction.copy()
        line_dest.x *= distance
        line_dest.y *= distance
        line_dest.z *= distance
        line_dest = line_dest + origin
        if not data.obj:
            render.drawLine(
                origin,
                line_dest,
                [1, 0, 0, 1]
            )
        else:
            render.drawLine(
                origin,
                point,
                [0, 1, 0, 1]
            )
            render.drawLine(
                point,
                line_dest,
                [1, 0, 0, 1]
            )
    return data


class RayCastFaceData(RayCastData):
    pass


def raycast_face(
    caster: GameObject,
    origin: Vector,
    dest: Vector,
    distance: float = 0,
    prop: str = '',
    material: str = '',
    exclude: bool = False,
    xray: bool = False,
    local: bool = False,
    mask: int = 65535,
    face_data: bool = False,
    visualize: bool = False
) -> RayCastFaceData[GameObject, Vector, Vector, Vector, KX_PolyProxy, Vector]:
    """[DEPRECATED]\n
    Raycast from any point to any target. Returns additional face data.

    :param caster: casting object, this object will be ignored by the ray.
    :param origin: origin point; any vector or list.
    :param dest: target point; any vector or list.
    :param distance: distance the ray will be cast
    (0 means the ray will only be cast to target).
    :param prop: look only for objects with this property.
    :param material: look only for objects with this material applied.
    :param exclude: invert the selection for `prop` and `material`.
    :param xray: look for objects behind others.
    :param local: add the target vector to the origin.
    :param mask: Collision Mask for this ray.
    :param visualize: show the raycast.

    :returns: (`obj`, `point`, `normal`, `direction`, `face`, `uv`)
    """
    console.warning("'uplogic.utils.raycasting.raycast_face()' is deprecated, use '...raycasting.raycast(face_data=True)' instead")
    return RayCastFaceData((None, None, None, None, None, None))


class RayCastDataProjectile(RayCastData):

    @property
    def points(self) -> list[Vector]:
        return self[3]

    @property
    def direction(self) -> Vector:
        p = self.points
        if len(p) > 1:
            return (p[-1] - p[-2]).normalized()
        else:
            return Vector((0, 0, 0))


def raycast_projectile(
    caster: GameObject,
    origin: Vector,
    aim: Vector,
    power: float,
    distance: float = 100,
    resolution: float = .05,
    prop: str = '',
    material: str = '',
    xray: bool = False,
    local: bool = False,
    mask: int = 65535,
    gravity: Vector = None,
    face_data: bool = False,
    visualize: bool = False
) -> RayCastDataProjectile[GameObject, Vector, Vector, Vector, KX_PolyProxy, Vector, list]:
    """Raycast along the predicted parabola of a projectile.

    :param caster: casting object, this object will be ignored by the ray.
    :param origin: origin point; any vector or list.
    :param aim: target point; the parabola will start towards this point.
    :param power: "speed" of the projectile; a higher values mean further throws
    :param distance: total distance the ray will be cast
    :param resolution: detail quality of the parabola; higher values mean less detail
    :param prop: look only for objects with this property.
    :param material: look only for objects with this material.
    :param xray: look for objects behind others.
    :param local: add the target vector to the origin.
    :param mask: Collision Mask for this ray.
    :param gravity: Define custom gravity. If left at `None`, scene gravity is used.
    :param face_data: Include face data in the returned object. Comes at a slight extra cost.
    :param visualize: show the raycast.

    :returns: (`obj`, `point`, `normal`, `points`)
    """
    def calc_projectile(t, vel, pos, gravity):
        half: float = gravity * (.5 * t * t)
        vel = vel * t
        return half + vel + pos

    if not local:
        aim = aim - origin
    aim.normalize()
    aim *= power
    origin = getattr(origin, 'worldPosition', origin)

    points: list = [origin]
    color: list = [1, 0, 0]
    idx = 0
    total_dist: float = 0
    resolution = clamp(resolution, .01, .99)

    grav = gravity if gravity else logic.getCurrentScene().gravity
    while total_dist < distance:
        target = (calc_projectile(idx, aim, origin, grav))
        start = origin if not points else points[-1]
        data = raycast(
            caster=caster,
            origin=start,
            dest=target,
            prop=prop,
            material=material,
            xray=xray,
            local=False,
            mask=mask,
            face_data=face_data,
            visualize=False
        )
        total_dist += (target-start).length
        if not data.obj:
            points.append(target)
        else:
            points.append(data.point)
            color = [0, 1, 0]
            break
        idx += resolution
    if visualize:
        for i, p in enumerate(points):
            if i < len(points) - 1:
                render.drawLine(p, points[i+1], color)
    return RayCastDataProjectile((data.obj, data.point, data.normal, points, data.face, data.uv))


class RayCastCameraData(tuple):
    @property
    def obj(self) -> GameObject:
        return self[0]

    @property
    def point(self) -> Vector:
        return self[1]

    @property
    def normal(self) -> Vector:
        return self[2]


def raycast_camera(
    distance: float = 0,
    prop: str = '',
    xray: bool = False,
    aim: Vector = Vector((.5, .5)),
    mask: int = 65535
) -> RayCastCameraData:
    """
    [DEPRECATED]

    Raycast from any point to any target. Returns additional face data.

    :param distance: distance the ray will be cast
    :param prop: look only for objects with this property.
    :param xray: look for objects behind others.
    :param aim: X and Y coordinates of the screen from 0-1
    :param mask: Collision Mask for this ray.

    :returns: (`obj`, `point`, `normal`)
    """
    # assume screen coordinates
    from ..console import warning
    warning("'raycasting.raycast_camera' is deprecated and will be removed in future versions, please use 'raycasting.raycast_screen' instead!")
    camera = logic.getCurrentScene().active_camera
    if isinstance(aim, Vector) and len(aim) == 2:
        vec = 10 * camera.getScreenVect(aim[0], aim[1])
        ray_target = camera.worldPosition - vec
        aim = ray_target
    obj, point, normal = camera.rayCast(
        aim,
        None,
        distance,
        prop,
        xray=xray,
        mask=mask
    )
    return RayCastCameraData((obj, point, normal))


def raycast_screen(
    caster=None,
    aim: Vector = None,
    distance: float = 100,
    prop: str = '',
    material: str = '',
    xray: bool = False,
    mask: int = 65535,
    face_data: bool = False

) -> RayCastCameraData[GameObject, Vector, Vector, Vector, KX_PolyProxy, Vector]:
    """Raycast from any point to any target. Returns additional face data.

    :param caster: Caster object, this object will be ignored by the raycast itself
    :param aim: Target screen coordinates
    :param distance: look for objects behind others.
    :param prop: Only look for objects that have this property attached
    :param material: Only look for objects that have this material attached
    :param xray: Ignore objects that don't have either material or property attached
    :param mask: Collision Mask for this ray.
    :param face_data: Add `KX_Polygon` and `Vector` of hitpoint UV coordinates to result.

    :returns: (`obj`, `point`, `normal`, `direction`, `polygon`, `uv`)
    """
    # assume screen coordinates
    camera = logic.getCurrentScene().active_camera
    if isinstance(aim, Vector) and len(aim) == 2:
        vec = 10 * camera.getScreenVect(aim[0], aim[1])
    else:
        mpos = logic.mouse.position
        vec = 10 * camera.getScreenVect(*mpos)
    ray_target = camera.worldPosition - vec
    aim = ray_target
    data = raycast(
        caster=caster if caster is not None else camera,
        dest=aim,
        origin=camera.worldPosition,
        distance=distance,
        material=material,
        prop=prop,
        xray=xray,
        mask=mask,
        face_data=face_data
    )
    return RayCastData(data)


def raycast_mouse(
    distance: float = 100,
    prop: str = '',
    material: str = '',
    exclude: bool = False,
    xray: bool = False,
    mask: int = 65535
) -> RayCastData:
    """
    [DEPRECATED]

    Raycast from the active camera to world cursor coordinates.

    :param distance: distance the ray will be cast
    :param prop: look only for objects with this property.
    :param material: look only for objects with this material applied.
    :param exclude: invert the selection for `prop` and `material`.
    :param xray: look for objects behind others.
    :param mask: Collision Mask for this ray.

    :returns: (`obj`, `point`, `normal`, `direction`, `None`, `None`)
    """
    from ..console import warning
    warning("'raycasting.raycast_camera' is deprecated and will be removed in future versions, please use 'raycasting.raycast_screen' instead!")
    camera = logic.getCurrentScene().active_camera
    mpos = logic.mouse.position
    vec = 10 * camera.getScreenVect(*mpos)
    ray_target = camera.worldPosition - vec
    return raycast(
        camera,
        camera.worldPosition,
        ray_target,
        distance,
        prop,
        material,
        exclude,
        xray,
        mask=mask
    )



import bpy
import gpu
# from gpu_extras.batch import batch_for_shader
from mathutils import Matrix
from mathutils import Vector
from random import uniform, gauss, random
from uplogic import utils
import bge
from uplogic.events import schedule
import time


def batch_for_shader(shader, type, content, *, indices=None):
    """
    Return a batch already configured and compatible with the shader.

    :arg shader: shader for which a compatible format will be computed.
    :type shader: :class:`gpu.types.GPUShader`
    :arg type: "'POINTS', 'LINES', 'TRIS' or 'LINES_ADJ'".
    :type type: str
    :arg content: Maps the name of the shader attribute with the data to fill the vertex buffer.
    :type content: dict
    :return: compatible batch
    :rtype: :class:`gpu.types.GPUBatch`
    """
    from gpu.types import (
        GPUBatch,
        GPUIndexBuf,
        GPUVertBuf,
        GPUVertFormat,
    )

    def recommended_comp_type(attr_type):
        if attr_type in {'FLOAT', 'VEC2', 'VEC3', 'VEC4', 'MAT3', 'MAT4'}:
            return 'F32'
        if attr_type in {'UINT', 'UVEC2', 'UVEC3', 'UVEC4'}:
            return 'U32'
        return 'I32'

    def recommended_attr_len(attr_name):
        attr_len = 1
        try:
            item = content[attr_name][0]
            while True:
                attr_len *= len(item)
                item = item[0]
        except (TypeError, IndexError):
            pass
        return attr_len

    def recommended_fetch_mode(comp_type):
        if comp_type == 'F32':
            return 'FLOAT'
        return 'INT'

    for data in content.values():
        vbo_len = len(data)
        break
    else:
        raise ValueError("Empty 'content'")

    vbo_format = GPUVertFormat()
    attrs_info = shader.attrs_info_get()
    for name, attr_type in attrs_info:
        comp_type = recommended_comp_type(attr_type)
        attr_len = recommended_attr_len(name)
        vbo_format.attr_add(id=name, comp_type=comp_type, len=attr_len, fetch_mode=recommended_fetch_mode(comp_type))

    vbo = GPUVertBuf(vbo_format, vbo_len)

    for id, data in content.items():
        if len(data) != vbo_len:
            raise ValueError("Length mismatch for 'content' values")
        vbo.attr_fill(id, data)

    if indices is None:
        return GPUBatch(type=type, buf=vbo)
    else:
        ibo = GPUIndexBuf(type=type, seq=indices)
        return GPUBatch(type=type, buf=vbo, elem=ibo)


class Particle:
    
    def __init__(self, system):
        self.system = system

        vert_out = gpu.types.GPUStageInterfaceInfo("my_interface")
        vert_out.smooth('VEC3', "pos")

        self.vertex_shader = """
        
        in vec3 position;
        
        uniform mat4 viewProjectionMatrix;
        uniform vec3 particle_transform;

        out vec3 pos;

        void main()
        {
            vec4 tpos = vec4(particle_transform + position, 1.0f);
            gl_Position = viewProjectionMatrix * tpos;
        }
        """

        self.fragment_shader = """
        out vec4 fragColor;

        void main()
        {
            fragColor = vec4(0.7);
        }
        """

        self.shader = gpu.types.GPUShader(self.vertex_shader, self.fragment_shader)
        self.shader.bind()

        # self.shader = gpu.shader.create_from_info(shader_info)
        # del vert_out
        # del shader_info

        scale = uniform(system.size - system.size_random, system.size + system.size_random) * .5
        self.life_random = uniform(-system.life, system.life)
        
        self.change_velocity()
        self.velocity = self.target_velocity
        # self.velocity = Vector((
        #     uniform(system.velocity[0] - system.velocity_random[0], system.velocity[0] + system.velocity_random[0]),
        #     uniform(system.velocity[1] - system.velocity_random[1], system.velocity[1] + system.velocity_random[1]),
        #     uniform(system.velocity[2] - system.velocity_random[2], system.velocity[2] + system.velocity_random[2])
        # ))

        self.init_pos = self.system.location + Vector((
            uniform(system.start_pos[0] - system.start_pos_random[0], system.start_pos[0] + system.start_pos_random[0]),
            uniform(system.start_pos[1] - system.start_pos_random[1], system.start_pos[1] + system.start_pos_random[1]),
            uniform(system.start_pos[2] - system.start_pos_random[2], system.start_pos[2] + system.start_pos_random[2])
        ))

        coords = [
            # Vector((scale, scale, scale)),
            # Vector((scale, -scale, -scale)),
            # Vector((-scale, scale, scale)),
            # Vector((-scale, scale, -scale)),
            # Vector((scale, scale, scale)),
            Vector((uniform(-scale, scale), uniform(-scale, scale), uniform(-scale, scale))),
            Vector((uniform(-scale, scale), uniform(-scale, scale), uniform(-scale, scale))),
            Vector((uniform(-scale, scale), uniform(-scale, scale), uniform(-scale, scale))),
            Vector((uniform(-scale, scale), uniform(-scale, scale), uniform(-scale, scale))),
            Vector((uniform(-scale, scale), uniform(-scale, scale), uniform(-scale, scale))),
            Vector((uniform(-scale, scale), uniform(-scale, scale), uniform(-scale, scale))),
            Vector((uniform(-scale, scale), uniform(-scale, scale), uniform(-scale, scale))),
            Vector((uniform(-scale, scale), uniform(-scale, scale), uniform(-scale, scale))),
            Vector((uniform(-scale, scale), uniform(-scale, scale), uniform(-scale, scale))),
            Vector((uniform(-scale, scale), uniform(-scale, scale), uniform(-scale, scale))),
            Vector((uniform(-scale, scale), uniform(-scale, scale), uniform(-scale, scale)))
        ]
        self.batch = batch_for_shader(self.shader, 'TRI_FAN', {'position': coords})
        self.pos = self.init_pos.copy()
        self.life = time.time()
        delay = uniform(0, system.life + self.life_random)
        self.stopped = True
        self.started = False
        self.spawn_chance = random()
        schedule(self.start, delay)

    def start(self):
        system = self.system
        if self.spawn_chance > system.emission_rate:
            return
        self.stopped = False
        self.started = True
        self.life = time.time()
        if system.use_gauss:
            self.init_pos = self.system.location + Vector((
                gauss(system.start_pos[0], system.start_pos_random[0] * .33),
                gauss(system.start_pos[1], system.start_pos_random[1] * .33),
                gauss(system.start_pos[2], system.start_pos_random[2] * .33)
            ))
        else:
            self.init_pos = self.system.location + Vector((
                uniform(system.start_pos[0] - system.start_pos_random[0], system.start_pos[0] + system.start_pos_random[0]),
                uniform(system.start_pos[1] - system.start_pos_random[1], system.start_pos[1] + system.start_pos_random[1]),
                uniform(system.start_pos[2] - system.start_pos_random[2], system.start_pos[2] + system.start_pos_random[2])
            ))
        self.life_random = uniform(-system.life, system.life)
        self.pos = self.init_pos.copy()

    def change_velocity(self):
        sys = self.system
        self.target_velocity = Vector((
            uniform(sys.velocity[0] - sys.velocity_random[0], sys.velocity[0] + sys.velocity_random[0]),
            uniform(sys.velocity[1] - sys.velocity_random[1], sys.velocity[1] + sys.velocity_random[1]),
            uniform(sys.velocity[2] - sys.velocity_random[2], sys.velocity[2] + sys.velocity_random[2])
        ))
        schedule(self.change_velocity, uniform(0, 1))

    def check_visible(self):
        if not self.started:
            return
        sys = self.system
        self.pos += self.velocity
        point = sys.cam.rayCast(sys.cam.worldPosition, self.pos)[0]
        if not point:
            self.draw()
    
    def calc_gravity(self):
        t = time.time() - self.life
        half: float = self.system.gravity * (.5 * t * t)
        vel = self.velocity * t
        return half + vel + self.pos

    def update(self):
        if self.stopped:
            return
        sys = self.system
        self.check_visible()
        self.velocity = self.velocity.lerp(self.target_velocity, .01)
        dat = sys.cam.rayCast(self.pos + self.velocity, self.pos)
        if dat[0]:
            self.pos = dat[1] + dat[2] * .01
            self.end()
        elif sys.time - self.life > sys.life + self.life_random:
            self.end()
    
    def end(self):
        self.stopped = True
        self.start()

    def draw(self):
        gpu.state.depth_test_set('ALWAYS')  # NONE, ALWAYS, GREATER and GREATER_EQUAL works, 
        gpu.state.depth_mask_set(True)
        matrix = bpy.context.region_data.perspective_matrix
        self.shader.uniform_float("viewProjectionMatrix", matrix)
        trans = self.pos
        self.shader.uniform_float("particle_transform", trans)
        self.batch.draw(self.shader)
        gpu.state.depth_mask_set(False)


class ParticleSystem:
    
    def __init__(
        self,
        amount=1000,
        life=10,
        life_random=0,
        start_pos=Vector((0, 0, 0)),
        start_pos_random=Vector((0, 0, 0)),
        size=.5,
        size_random=0.0,
        velocity=Vector((0, 0, 0)),
        velocity_random=Vector((0, 0, 0)),
        gravity=Vector((0, 0, -9.8)),
        location=Vector((0, 0, 0)),
        use_gauss=False,
        exclude_occlusion_mask=32768
    ):
        self.life = life
        self.life_random = life_random
        self.emission_rate = 1.0
        self.exclude_occlusion = 65535 - exclude_occlusion_mask
        self.particles = []
        self.start_pos = Vector(start_pos)
        self.start_pos_random = Vector(start_pos_random)
        self.size = size
        self.use_gauss = use_gauss
        self.size_random = size_random
        self.velocity = Vector(velocity)
        self.velocity_random = Vector(velocity_random)
        self.gravity = gravity

        self.location = Vector(location)
        
        self.time = time.time()
        
        for x in range(amount):
            self.particles.append(Particle(self))

        bge.logic.getCurrentScene().post_draw.append(self.update)
    
    def update(self):
        self.cam = bge.logic.getCurrentScene().active_camera
        self.time = time.time()
        for p in self.particles:
            if p.stopped:
                p.start()
            p.check_visible() if p.stopped else p.update()




from ..image import Image


class ImagePreview(Image):

    fragment_shader = """
    in vec2 uv;
    out vec4 fragColor;

    uniform sampler2D image;
    uniform float alpha = 1.0;

    void main() {
        vec4 color = mix(vec4(0.0), texture(image, uv), alpha);
        fragColor = pow(color, vec4(1));
    }
    """


from uplogic.nodes import ULConditionNode


class ULLogicGate(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.ca = None
        self.cb = None
        self.OUT = self.add_output(self.get_and)

        self.getters = [
            self.get_and,
            self.get_or,
            self.get_xor,
            self.get_not,
            self.get_nand,
            self.get_nor,
            self.get_xnor,
            self.get_and_not
        ]
        self.gate = 0

    @property
    def gate(self):
        return None

    @gate.setter
    def gate(self, val):
        self.OUT._value_getter = self.getters[val]

    def get_and(self):
        ca = self.get_condition(self.ca)
        return ca and self.get_condition(self.cb)

    def get_or(self):
        ca = self.get_condition(self.ca)
        cb = self.get_condition(self.cb)
        return ca or cb

    def get_xor(self):
        ca = self.get_condition(self.ca)
        cb = self.get_condition(self.cb)
        return (ca or cb) and not (ca and cb)

    def get_not(self):
        condition = self.get_condition(self.ca)
        return not condition

    def get_nand(self):
        ca = self.get_condition(self.ca)
        cb = self.get_condition(self.cb)
        return not (ca and cb)

    def get_nor(self):
        ca = self.get_condition(self.ca)
        cb = self.get_condition(self.cb)
        return not (ca or cb)

    def get_xnor(self):
        ca = self.get_condition(self.ca)
        cb = self.get_condition(self.cb)
        return not (ca or cb) or (ca and cb)

    def get_and_not(self):
        ca = self.get_condition(self.ca)
        cb = self.get_condition(self.cb)
        return ca and not cb



from uplogic.nodes import ULParameterNode
from bge.logic import getCurrentScene


class GetObjectByNameNode(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.object_name = None
        self.GAME_OBJECT = self.add_output(self.get_object)

    def get_object(self):
        object_name: str = self.get_input(self.object_name)
        return getCurrentScene().objects.get(object_name, None)



from .shader import Filter2D


glsl = """
/* Original Code by existical @shadertoy: https://www.shadertoy.com/view/Xltfzj
Modified by Iza Zed for UPBGE
*/

uniform sampler2D bgl_RenderedTexture;
uniform float bgl_RenderedTextureWidth;
uniform float bgl_RenderedTextureHeight;
vec2 resolution = vec2(bgl_RenderedTextureWidth, bgl_RenderedTextureHeight);

// out vec4 fragColor;
// in vec4 bgl_TexCoord;

vec2 texcoord = bgl_TexCoord.xy;

uniform float samples;
uniform float power;

void main()
{
    float Pi = 6.28318530718;

    float quality = 9.0;
   
    vec2 radius = power / resolution.xy;
    
    vec2 uv = texcoord;

    vec4 color = texture(bgl_RenderedTexture, uv);
    
    // Blur calculations
    for (float d = 0.0; d < Pi; d += Pi / samples)
    {
		for(float i = 1.0 / quality; i <= 1.0; i += 1.0 / quality)
        {
			color += texture(bgl_RenderedTexture, uv + vec2(cos(d), sin(d)) * radius * i);
        }
    }
    color /= quality;
    fragColor = color;
}"""


class Blur(Filter2D):

    def __init__(self, power=1.0, samples=16, idx: int = None) -> None:
       self.uniforms = {'samples': float(samples), 'power': float(power)}
       super().__init__(glsl, idx, {'samples': self.uniforms, 'power': self.uniforms})

    @property
    def samples(self):
        return self.uniforms['samples']

    @samples.setter
    def samples(self, val):
        self.uniforms['samples'] = float(val)

    @property
    def power(self):
        return self.uniforms['power']

    @power.setter
    def power(self, val):
        self.uniforms['power'] = float(val)



from uplogic.nodes import ULConditionNode
from uplogic.utils.constants import LOGIC_OPERATORS


class ULCompareVectors(ULConditionNode):
    def __init__(self, operator='GREATER'):
        ULConditionNode.__init__(self)
        self.operator = operator
        self.all = None
        self.threshold = None
        self.param_a = None
        self.param_b = None
        self.OUT = self.add_output(self.get_result)

    def get_result(self):
        a = self.get_input(self.param_a)
        b = self.get_input(self.param_b)
        all_values = self.get_input(self.all)
        operator = self.get_input(self.operator)
        threshold = self.get_input(self.threshold)
        return self.get_vec_val(
            operator,
            a,
            b,
            all_values,
            threshold
        )

    def get_vec_val(self, op, a, b, xyz, threshold):
        for ax in ['x', 'y', 'z']:
            av = getattr(a, ax)
            bv = getattr(b, ax)
            av = bv if abs(av - bv) < threshold else av
            if xyz[ax] and not LOGIC_OPERATORS[op](av, bv):
                return False
        return True



from uplogic.nodes import ULActionNode
from uplogic.utils.objects import Curve
from uplogic.utils import draw_cube


class DistributeCurvePointsNode(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = False
        self.curve = None
        self.debug = False
        self.custom_density = False
        self.density = 1
        self._curve = None
        self._points = None
        self.DONE = self.add_output(self.get_done)
        self.POINTS = self.add_output(self.get_points)

    def get_done(self):
        return self._done

    def get_points(self):
        return self._points

    def distribute_points(self):
        len_points = max(len(self._curve.points), 1)
        res = self.density if self.custom_density else self._curve.resolution * len_points
        fac = 1 / (res)
        return [self._curve.evaluate(x * fac) for x in range(res + 1)]

    def evaluate(self):
        if self.get_condition() or self._curve is None:
            self._curve = Curve(self.get_input(self.curve), use_evaluate=True)
            self._points = self.distribute_points()
            self._done = True
        if self.debug:
            for x in self._points:
                draw_cube(x, centered=True, width=.2)
